"""Load default YAML settings file, then look for user defined settings and create a ChainMap."""
from collections.abc import MutableMapping
from collections import defaultdict
from argparse import ArgumentParser
from pathlib import Path

import functools
import inspect
import yaml
import os


class ConfigurationError(Exception):
    pass


try:
    from yamlinclude import YamlIncludeConstructor
except ImportError:
    YamlIncludeConstructor = None

from argus_cli.helpers import log as log_module


# FIXME: Legacy. This instance should be removed, but is depended by api_generator.
ARGUS_CLI_CONFIG_LOCATION = Path(
    os.environ.get("ARGUS_CLI_SETTINGS", os.path.expanduser("~/.argus_cli.yaml"))
)
ARGUS_CLI_SETTINGS_INCLUDES = Path(
    os.environ.get("ARGUS_CLI_SETTINGS_INCLUDES", "~/.argus_cli.d")
).expanduser()

MISSING_INCLUDES_ERROR = "could not determine a constructor for the tag '!include'"
NO_INCLUDES_MSG = "the configuration uses includes but they are not enabled - install pyyaml-include to enable includes"

if YamlIncludeConstructor:
    # setup yaml includes if available
    YamlIncludeConstructor.add_to_loader_class(
        loader_class=yaml.SafeLoader, base_dir=ARGUS_CLI_SETTINGS_INCLUDES
    )


def _merge(source: dict, destination: dict) -> dict:
    """Deep merges two dictionaries and replaces values from destination with the ones from source."""
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            _merge(value, node)
        else:
            destination[key] = value

    return destination


def _get_environment() -> str:
    # TODO: This should really be handled in a better way, but might need some refactoring
    provider = ArgumentParser(add_help=False)
    provider.add_argument("--environment")
    arguments = vars(provider.parse_known_args()[0])

    return arguments.get("environment")


def _get_debug_mode(config: dict) -> dict:
    """Sets up debug mode if there is a --debug argument on the commandline"""
    provider = ArgumentParser(add_help=False)
    provider.add_argument("--debug", action="store_true")
    arguments = vars(provider.parse_known_args()[0])

    if arguments.get("debug"):
        config["global"] = {"debug": True}

        for logger in config["logging"]["handlers"].values():
            logger["level"] = "DEBUG"

        log_module.log.info("Debug mode activated!")

    return config


def nested_defaultdict():
    return defaultdict(nested_defaultdict)


class Config(MutableMapping):
    """Loads and stores data from configuration.

    Configuration is either loaded from the system environment variables or
    the provided configuration file.

    Environments
    ------------

    The configuration files can also have a custom "environments" inside them,
    for specific use-cases. For example when a user wants to run a command in a
    development environment.

    These custom environments will be found under the top level key
    `environments`. Under this key the first key will be the name of the
    custom environment, and bellow this again will be the normal keys.

    In the following example, we have a development environment with a
    custom URL.

    .. code-block:: yaml
        api:
            url: https://example.com/

        environments:
            development:
                api:
                    url: https://test.example.com/
    """

    #: Stores the configuration variables
    _dict = nested_defaultdict()

    env_key = None
    config_key = None

    def __init__(
        self,
        env_prefix: str = None,
        config_file: Path = None,
        base_config: dict = None,
        environment: str = None,
    ):
        """Creates a new config

        :param env_prefix: Will be prepended to all env checks (e.g. "<PREFIX>_KEY")
        :param config_file: The file where the config can be found
        :param base_config: Config base to base off
        :param environment: Environment tag in the settings to load on get
        """
        if not base_config:
            base_config = {}

        if config_file:
            with open(str(config_file.absolute())) as f:
                try:
                    file_config = yaml.safe_load(f) or {}
                except yaml.constructor.ConstructorError as e:
                    if MISSING_INCLUDES_ERROR in e.problem:
                        raise ConfigurationError(NO_INCLUDES_MSG)
                    raise

        else:
            file_config = {}

        self.env_prefix = (
            env_prefix or ""
        )  # With no prefix, it can be left blank as we're doing string formatting.

        self._dict = _merge(file_config, base_config)
        if environment:
            try:
                self._dict = _merge(
                    # If a key is empty, it will be None. That would make _merge raise an exception.
                    self._dict["environments"][environment] or {},
                    self._dict,
                )
            except KeyError as e:
                # Give the key error a bit more descriptive text
                raise KeyError(
                    "Could not find custom environments '{env}'. Is the environment"
                    " present under 'environments' in your config file?".format(
                        env=environment
                    )
                )

    def __call__(
        self, func: callable = None, *, env_key: str = None, config_key: str = None
    ):
        """Decorator to create a partial function with arguments from config or env filled in.

        Can be called with or without parameters.

        Example:
            >>> @config
            Or:
            >>> @config(env_key="something", config_key="funny")

        :param func: Function to decorate
        :param env_key: Environment key for this function
        :param config_key: Config key for this function
        """
        if not func:
            return functools.partial(self, config_key=config_key, env_key=env_key)

        # TODO: Find a cleaner way to do this
        self.config_key = config_key
        self.env_key = env_key

        arguments = {
            name: self[name]
            for name in inspect.signature(func).parameters.keys()
            if name in self
        }

        self.env_key = None
        self.config_key = None

        return functools.partial(func, **arguments)

    def __getitem__(self, key):
        """Gets an item from either the dict or the env."""
        env_key = key
        if self.env_prefix:
            env_key = self.env_prefix + "_" + key
        if self.env_key:
            env_key = self.env_key + "_" + key

        try:
            # Try to get from the environment
            return os.environ[env_key.upper()]
        except KeyError:
            if self.config_key:
                return self._dict[self.config_key][key]
            return self._dict[key]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)


def generate_config(environment: str = None):
    try:
        base_config = yaml.safe_load(
            open(Path(__file__).resolve().parent / "resources" / "config.yaml")
        )
    except yaml.constructor.ConstructorError as e:
        if MISSING_INCLUDES_ERROR in e.problem:
            raise ConfigurationError(NO_INCLUDES_MSG)
        raise
    return Config(
        env_prefix="ARGUS",
        config_file=ARGUS_CLI_CONFIG_LOCATION
        if ARGUS_CLI_CONFIG_LOCATION.exists()
        else None,
        environment=environment,
        base_config=base_config,
    )


settings = generate_config(_get_environment())
_get_debug_mode(settings)
log_module.setup_logger(settings["logging"])



import importlib
import pkgutil

from argus_cli import register_provider
from argus_cli.settings import settings
from argus_cli.helpers.log import log


# TODO: Debug isn't used, but rather look at settings._get_debug_mode.
#       This should be fixed.
# TODO: Environment isn't used, same thing as with --debug.
@register_provider()
def argus_api(apikey: str = None, debug: bool = False, environment: str = None) -> None:
    """Argus CLI provider"""
    if apikey:
        settings["api"]["api_key"] = apikey


def import_submodules(
    package: str, exclude_name: str = None, recursive: bool = True
) -> dict:
    """Import all submodules of a module, recursively.

    This is used to import all APIs when Argus is loaded,
    so that the commands become registered as plugins,
    but can also be used to recursively import any other
    package where you want every single file to load.

    :param package: Package name, e.g "argus_api.api"
    :param exclude_name: Any module containing this string will not be imported
    :param recursive: Recourse trough the package structure to import files?
    """
    module = importlib.import_module(package)

    results = {}

    for loader, name, is_pkg in pkgutil.iter_modules(
        module.__path__, prefix=package + "."
    ):
        if exclude_name and exclude_name in name:
            log.debug(f"{name} is in exclude_name, ignoring")
            continue

        results[name] = importlib.import_module(name)
        if recursive and is_pkg:
            results.update(import_submodules(name, exclude_name=exclude_name))
    return results



"""
Provides enrichments: functions that will add data to the input objects.

all enrichment data should be placed under a top-level field in order to avoid
name collisions with API fields.
"""
import enum
from typing import Union, Iterable
from functools import lru_cache

from argus_api.api.customernetworks.v1.network import search_customer_network
from argus_cli.helpers.log import log

# enrichment data will be stored in the following field of the enriched object's
# dictionary.
ENRICHMENTS_FIELD_NAME = "_enrichments_"


class Enrichments(enum.Enum):
    """Valid enrichment names"""

    EVENTS_CUSTOMER_NEWORKS_SOURCE = "customer-networks-source"
    EVENTS_CUSTOMER_NEWORKS_DESTINATION = "customer-networks-destination"

    @classmethod
    def has_value(cls, value):
        return value in (i.value for i in cls)


# enrichment applicable to event data
EVENTS_ENRICHMENTS = [
    Enrichments.EVENTS_CUSTOMER_NEWORKS_SOURCE.value,
    Enrichments.EVENTS_CUSTOMER_NEWORKS_DESTINATION.value,
]


class IPTypeField(enum.Enum):
    """Valid IP fields

    Event field from which to read - and enrichment field to write -
    IP address information from/to.
    """

    SOURCE = "source"
    DESTINATION = "destination"


def enrich_data(data: Iterable[dict], enrichments: Iterable[str]) -> None:
    """Performs the selected enrichment on the data

    :param data: input data as an iterable of API objects
    :param enrichments: names of enrichments to perform
    """
    for enrichment in (e for e in enrichments if not Enrichments.has_value(e)):
        log.warn(f'"{enrichment}" is not a valid enrichment name')

    if Enrichments.EVENTS_CUSTOMER_NEWORKS_SOURCE.value in enrichments:
        events_customer_networks(data, IPTypeField.SOURCE)
    if Enrichments.EVENTS_CUSTOMER_NEWORKS_DESTINATION.value in enrichments:
        events_customer_networks(data, IPTypeField.DESTINATION)


def events_customer_networks(events: Iterable[dict], field: IPTypeField) -> None:
    """Adds customer network information to events.

    Each event's IP address (source or destination, based on the value of the
    field parameter) is read from the event and searched for through the
    customer networks API.

    If the search has results, they are added to the enrichment field of the
    event as a list of API customer networks dictionaries. Otherwise, the
    field's value is set to None.

    The added fields are:
    event[ENRICHMENTS_FIELD_NAME]["customer-networks"][field]

    :param events: events to enrich
    :param field: the ip field from which the ip address will be read and
        destination field to which enrichment data will be writen.
    """

    enrich_field_name = "customer-networks"
    log.debug(f"runnning enrichment {enrich_field_name}-{field.value}")

    def enrich_event(event: dict) -> None:
        """Adds enrichment data to the event

        Calls the other nested functions to fetch enrichment data and adds it
        to an event.

        :param event: input event dict
        """
        ip = _get_ip(event)
        customer_id = _get_customer_id(event)
        enrich_field = event.setdefault(ENRICHMENTS_FIELD_NAME, {})
        enrich_subfield = enrich_field.setdefault(enrich_field_name, {})
        enrich_subfield.update({field.value: _lookup_networks(ip, customer_id)})

    def _get_ip(event: dict) -> Union[str, None]:
        """Get an event's source/destination IP address.

        :param event: input event dict
        :return: event's IP address or None
        """
        try:
            return event[field.value]["networkAddress"]["address"]
        except (KeyError, TypeError):
            return None

    def _get_customer_id(event: dict) -> Union[int, None]:
        """Get an event's customer ID.

        :param event: input event dict
        :return: event's IP customer ID or None
        """
        id_str = event.get("customerInfo", {}).get("id")
        return int(id_str) if id_str is not None else None

    def _get_net_mask(network: dict) -> Union[int, None]:
        """Extracts the mask from a Customer Network

        :param network: input custommer network's dict
        :return: network mask bits or None
        """
        return network.get("networkAddress", {}).get("maskBits")

    @lru_cache(maxsize=0)
    def _lookup_networks(ip: str, customerID: int) -> Union[dict, None]:
        """Returns customer networks for a given IP.

        :param ip: the IP to search for
        :param customerID: the customer's ID
        :return: list of customer networks dict or None if no results
        """
        if ip is None or customerID is None:
            return None
        networks = search_customer_network(customerID=customerID, addresses=[ip])
        networks_data = networks.get("data")
        if networks_data is not None:
            # sort from most specific to least specific
            networks_data.sort(key=_get_net_mask, reverse=True)
        return networks_data

    for event in events:
        enrich_event(event)



from argus_api.api.events.v1.aggregated import find_aggregated_event_stats

from argus_cli.plugin import register_command
from argus_cli.helpers.log import log
from argus_cli.utils import time
from argus_cli.utils.formatting import get_data_formatter
from argus_plugins import argus_cli_module

from argus_plugins.cases.utils import get_customer_id

STATS_GROUPINGS = [
    "ATTACK_IDENTIFIER",
    "ATTACK_CATEGORY",
    "ALARM",
    "LOCATION",
    "EVENT_SEVERITY",
    "SOURCEIP",
    "DESTINATIONIP",
    "PROTOCOL",
    "CUSTOMER",
    "SOURCE_GEOIP_LOCATION",
    "SOURCE_GEOIP_COUNTRY",
    "DESTINATION_GEOIP_LOCATION",
    "DESTINATION_GEOIP_COUNTRY",
]


def setup_fields(keys: list):
    headers = ["time", "count"]

    # Add the keys sent by Argus to the headers
    headers.extend([key for key in keys[0].keys() if key != "id"])

    return headers, keys


def create_presentable_data(time_frames: list, keys: list) -> list:
    """Creates presentable data from the returned statistics"""
    results = []

    for time_frame in time_frames:
        for record in time_frame["records"]:
            # Add constant data
            result = {
                "time": time.timestamp_to_date(
                    time_frame["timeFrame"]["startTime"] / 1000
                ),
                "count": record["value"]["count"],
            }

            # Add variable data
            for key, value in keys[record["keyID"] - 1].items():
                if key == "id":
                    # We don't want the ID field in our data
                    continue
                result[key] = value

            results.append(result)

    return results


@register_command(extending="events", module=argus_cli_module)
def statistics(
    start: time.date_or_relative,
    end: time.date_or_relative,
    customer: get_customer_id,
    resolution: time.time_diff = 0,
    group_by: STATS_GROUPINGS = None,
    attack_category: int = None,
    count_raw_events: bool = False,
    format: get_data_formatter = "csv",
):
    """Prints statistical data about events with some specified criteria

    :param start: Start time of the search
    :param end: End time of the search
    :param list customer: Customers to search for
    :param resolution: The timeframe of results
    :param list group_by: Parameters to group by
    :param list attack_category: Attack category to display
    :param count_raw_events: Count the raw number of events instead of aggregated events
    :param format: How to format the output. Can be either jira-table, csv or a formatted string.
    """
    log.info("Getting events from argus...")
    data = find_aggregated_event_stats(
        startTimestamp=start,
        endTimestamp=end,
        customerID=customer,
        resolution=resolution,
        countRawEvents=count_raw_events,
        attackCategoryID=attack_category,
        groupBy=group_by,
    )["data"]

    headers, keys = setup_fields(data["keys"])

    log.info("Creating presentable data...")
    presentable_data = create_presentable_data(data["timeFrames"], keys)

    log.info("Printing statistics...")
    print(format(presentable_data, headers=headers))



import keyword
from re import sub, match

from colorama import Fore, Style

from terminaltables import AsciiTable
from string import capwords


def to_caterpillar_case(camel_cased_string: str) -> str:
    """Replace UpperCase with upper-case and upperCase_Name with upper-case-name and splits into the dictionary levels

    :param camel_cased_string: CamelCasedString to convert
    :returns: caterpillar-cased-string
    """
    if match(r".*[a-z](ID).*", camel_cased_string):
        camel_cased_string = camel_cased_string.replace("ID", "Id")
    return (
        sub("(?<!^)((?=[A-Z])|\_)", "-", camel_cased_string)
        .replace("_", "-")
        .lower()
        .split(".")[-1]
    )


def to_snake_case(camel_cased_string: str) -> str:
    """Replace UpperCase with upper-case and upperCase_Name with upper_case_name and splits into the dictionary levels

    :param camel_cased_string: CamelCasedString to convert
    :returns: snake_cased_string
    """
    if match(r".*[a-z](ID).*", camel_cased_string):
        camel_cased_string = camel_cased_string.replace("ID", "Id")
    return sub("(?<!^)((?=[A-Z])|\-)", "_", camel_cased_string).lower().split(".")[-1]


def python_name_for(javascript_type: str) -> str:
    """Find the Python name for a javascript type

    :param str javascript_type: JavaScript type name, e.g array, string, integer
    :returns: Name for the type in python
    :rtype: str
    """
    alternatives = {
        "integer": "int",
        "number": "float",
        "float": "float",
        "object": "dict",
        "array": "list",
        "bool": "bool",
        "boolean": "bool",
        "string": "str",
        "null": "None",
    }

    if javascript_type in alternatives:
        return alternatives[javascript_type.lower()]
    else:
        return javascript_type


def to_safe_name(argument: str) -> str:
    """Python prevents us from using the built-in keywords as parameter names in
    functions, so if the API requests any of these keywords, we escape them with
    an underscore and provide a method to escape and unescape them

    :param str argument: Argument name to escape, if necessary
    :returns: _argument if the argument was unsafe
    """
    if keyword.iskeyword(argument):
        return "_%s" % argument
    return argument


def from_safe_name(argument: str) -> str:
    """Python prevents us from using the built-in keywords as parameter names in
    functions, so if the API requests any of these keywords, we escape them with
    an underscore and provide a method to escape and unescape them

    :param str argument: Argument name to escape, if necessary
    :returns: _argument if the argument was unsafe
    """
    if argument.startswith("_") and keyword.iskeyword(argument.replace("_", "")):
        return argument.replace("_", "")
    return argument


def clear() -> None:
    """Clears the terminal"""
    print("\033[H\033[J")


def ask(question: str) -> bool:
    """Prompts user with a a yes/no terminal prompt and returns True if the user answered yes

    :param bool question:
    :return:
    """
    yN = input("%s (Y/n) " % question)
    return yN.strip().lower() == "y"


def table(data: iter, keys: iter, format: callable = None, title: str = None) -> str:
    """Accepts a list of dicts or objects, and keys to extract from it, then creates a
    table that can be printed to the terminal.

    :param list data: Dataset to print as a table
    :param list keys: Keys to extract from each dict or object
    :param callable format: Optional formatter, will be called as format(value, key)
    :param str title: Title to print over the table (optional)
    :returns: str
    """
    # Print updates and let user accept before performing updates
    table_data = [
        list(map(capwords, keys)),
    ]

    if not format:
        format = lambda value, key: value

    for row in data:
        table_data.append(
            [
                format(getattr(row, key, None) or row.get(key, None), key=key)
                for key in keys
            ]
        )

    return AsciiTable(table_data, title=title if title else None).table


def red(text: str, **kwargs) -> str:
    return Fore.RED + text + Style.RESET_ALL


def yellow(text: str, **kwargs) -> str:
    return Fore.YELLOW + text + Style.RESET_ALL


def cyan(text: str, **kwargs) -> str:
    return Fore.CYAN + text + Style.RESET_ALL


def green(text: str, **kwargs) -> str:
    return Fore.GREEN + text + Style.RESET_ALL


def success(*args) -> str:
    return "%s %s" % (green("++"), " ".join(args))


def failure(*args) -> str:
    return "%s %s" % (red("--"), " ".join(args))



from typing import Optional, List, NamedTuple, Set
from datetime import datetime

from argus_api.api.cases.v2.case import (
    advanced_case_search,
    list_case_history,
    list_transaction_notifications,
)
from argus_api.api.cases.v2.servicesubscription import list_subscriptions

from argus_cli.utils.formatting import get_data_formatter, FORMATS
from argus_cli.helpers.log import log
from argus_cli.helpers.pagination import offset_paginated
from argus_cli.plugin import register_command
from argus_plugins import argus_cli_module
from argus_plugins.cases.utils import get_customer_id

MNEMONIC_CUSTOMER_ID = 1

#: services a customer must be subscribed to in order to be included in the check
FILTER_SERVICES = ["ids", "logstorage", "irt"]
#: services to which a case must belong to be included in the check
MONITORED_SERVICES = ["ids", "logstorage", "irt", "administrative"]


class SubscribedCustomer(NamedTuple):
    id: int
    name: str
    shortName: str


def get_service_customers() -> Set[SubscribedCustomer]:
    """returns the set of customers subscribed to services that are relevant for the
    notification check.
    """
    customers = set()
    for service in FILTER_SERVICES:
        for page in offset_paginated(list_subscriptions)(service=service):
            customers.update(
                SubscribedCustomer(
                    id=sub["customer"]["id"],
                    name=sub["customer"]["name"],
                    shortName=sub["customer"]["shortName"],
                )
                for sub in page["data"]
            )
    return customers


def customer_is_notified(case_id: int, transaction_id: str) -> bool:
    """Checks if the any of the contacts are NOT a mnemonic email."""
    notifications = list_transaction_notifications(
        case_id,
        transaction_id,
    )["data"]

    return any(
        notification["contact"].split("@")[-1] != "mnemonic.no"
        for notification in notifications
    )


def _is_internal(transaction_or_change: dict) -> bool:
    obj = transaction_or_change.get("object", {})
    if not isinstance(obj, dict):
        return False
    return "INTERNAL" in obj.get("flags", [])


def _is_status_change(transaction: dict) -> bool:
    """a transaction is a status change if only status and/or priority are listed in
    the changes.

    If an internal comment has been added in the transactions it is still only a status
    change from the perspective of the user, who can't see it.
    """
    changes = filter(lambda change: not _is_internal(change), transaction["changes"])
    return all(c["field"] in ("status", "priority", "assignedTechID") for c in changes)


def _is_from_mnemonic(case: dict, transaction: dict) -> bool:
    """Check whether a transaction originates from Mnemonic"""
    if transaction["operation"] == "createCase":
        return "SUBMITTED_BY_TECH" in case["flags"]
    else:
        return transaction["user"]["customer"].get("id", -1) == MNEMONIC_CUSTOMER_ID


def non_notified_transactions(
    case: dict, start: int, end: int, ignored_users=Optional[List[str]]
):
    """Returns all transactions where the customer wasn't notified and it wasn't an internal comment."""

    transactions = list_case_history(
        case["id"],
        operation=["createCase", "addCaseComment", "updateCase", "publishCase"],
        startTimestamp=start,
        endTimestamp=end,
    )["data"]

    for transaction in transactions:

        is_internal = _is_internal(transaction)
        skipped_notification = "SKIP_NOTIFICATION" in transaction.get("flags", [])
        is_status_change = _is_status_change(transaction)
        user_is_ignored = transaction["user"]["userName"] in ignored_users
        from_mnemonic = _is_from_mnemonic(case, transaction)

        if any(
            (
                is_internal,
                is_status_change,
                skipped_notification,
                not from_mnemonic,
                user_is_ignored,
            )
        ):
            # In the case that the comment is internal or is marked as a comment
            # that should be skipped: Ignore the given transaction.
            continue
        elif not customer_is_notified(case["id"], transaction["id"]):
            log.info(
                f"case {case['id']} ({case['customer']['shortName']}) : non notified transaction found : {transaction['id']}"
            )
            yield transaction


@register_command(extending="cases", module=argus_cli_module)
def notification_check(
    start: datetime,
    end: datetime,
    exclude_customer: get_customer_id = [],
    ignore_user: str = [],
    format: FORMATS = "csv",
):
    """Outputs all cases where the customer didn't get a notification

    :param start: The start of the search
    :param end:  The end of the search
    :param list exclude_customer: Customer (shortname) to exclude from the search. Mnemonic is automatically excluded
    :param list ignore_user: Users (shortNames) whose transactions will never be inspected for notifications
    :param format: How the output will be formatted.
    """
    log.warning(
        "This one-off script is going to be moved after ARGUS-11915 is completed!"
    )

    startTimestamp = int(start.timestamp() * 1000)
    endTimestamp = int(end.timestamp() * 1000)

    log.info("Fetching cases...")

    cases = advanced_case_search(
        startTimestamp=startTimestamp,
        endTimestamp=endTimestamp,
        includeDescription=False,
        excludeFlags=["INTERNAL"],
        service=MONITORED_SERVICES,
        customer=[customer.id for customer in get_service_customers()],
        subCriteria=[
            {
                "exclude": True,
                "customerID": [customer for customer in (exclude_customer + [1])],
            }
        ],
        limit=0,
    )["data"]

    log.info("Fetched {} cases".format(len(cases)))

    non_notified = []
    total = len(cases)
    count = 0
    for case in cases:
        count += 1
        log.info(f"Checking case #{case['id']} ({count}/{total})")
        if case["accessMode"] in ("readRestricted", "explicit"):
            log.debug(
                f"skipping case #{case['id']} due to access mode: {case['accessMode']}"
            )
            continue

        for transaction in non_notified_transactions(
            case,
            startTimestamp,
            endTimestamp,
            ignored_users=ignore_user,
        ):
            non_notified.append(
                {
                    "Case ID": "[{id}|https://argusweb.mnemonic.no/spa/case/view/{id}]".format(
                        id=case["id"]
                    ),
                    "Customer": case["customer"]["shortName"],
                    "Transaction Time": datetime.fromtimestamp(
                        transaction["timestamp"] / 1000
                    ).isoformat(sep=" "),
                    "User": transaction["user"]["userName"],
                    "Operation": transaction["operation"],
                    "Sub-Changes": [
                        change["field"] for change in transaction.get("changes", [])
                    ],
                }
            )

    print(get_data_formatter(format)(non_notified))



import difflib
import json
from collections import namedtuple

from argus_cli.helpers.log import log
from argus_cli.plugin import register_command
from argus_cli.utils import time, formatting
from argus_api.api.eventfilters.v1 import filter as filter_api
from argus_plugins import argus_cli_module

KIBANA_HISTORY_URL = "http://osl-analyze2:5601/#/discover?_g=(refreshInterval:(display:Off,pause:!f,section:0,value:0),time:(from:now-14d,mode:quick,to:now))&_a=(columns:!(_source),index:'aggrevent-*',interval:auto,query:(query_string:(analyze_wildcard:!t,query:'properties.argus_created_by_filterid:%s')),sort:!(startTime,desc))]) "
ARGUS_HISTORY_URL = "https://argusweb.mnemonic.no/spa/eventfilter/match/view/%s"

FILTER_TYPES = {
    "annotation": ["CustomEventAnnotationFilter"],
    "analysis": ["CustomEventAnalysisFilter"],
    "match": [
        "IPAttackEventMatchFilter",
        "NIDSEventMatchFilter",
        "AggregatedIPAttackEventMatchFilter",
    ],
}

REVISION_FIELDS = (
    "preStatement1",
    "preStatement2",
    "statementCode",
    "triggerCode",
)
UNWANTED_JSON_DIFF_FIELDS = (
    "lastUpdatedTimestamp",
    "lastUpdatedByUser",
    "revision",
    "masterID",
    "createdTimestamp",
    "flags",
    "statementCode",
    "triggerCode",
    "enabled",
    "preStatement1",
    "preStatement2",
    "index",
)

Filter = namedtuple(
    "Filter",
    ["id", "customer", "updated_by", "description", "enabled", "updated", "difference"],
)


def jira_output(data: list):
    """Creates output for JIRA"""
    out = ""

    for filt in data:
        if not filt.enabled:
            filter_status = "Disabled filter"
        elif filt.updated:
            filter_status = "Updated filter"
        else:
            filter_status = "New filter"

        filter_info = "%s ([Argus|%s]) ([History|%s})" % (
            filt.id,
            ARGUS_HISTORY_URL % filt.id,
            KIBANA_HISTORY_URL % filt.id,
        )

        out += (
            "h2. %s: %s\n"
            "*Filter:* %s\n"
            "*Customer:* %s\n"
            "*Updated by:* %s\n"
            % (filter_status, filt.description, filt.id, filt.customer, filt.updated_by)
        )

        for key, diff in filt.difference.items():
            out += "%s diff:\n{code}%s{code}" % (key, diff)

    return out


OUTPUT_METHODS = {
    "csv": lambda headers, data: formatting.csv(data, headers=headers),
    "jira": lambda headers, data: jira_output(
        data
    ),  # Creates a consistent "API" for output methods
}


def _diff(left: str, right: str, context: int = 0) -> str:
    """Wrapper around difflib.unified_diff"""
    return "".join(
        difflib.unified_diff(left.splitlines(True), right.splitlines(True), n=context)
    )


def _json_diff(left: dict, right: dict, context: int = 0) -> str:
    """Diffs a filter's JSON, and filters out some fields we don't want"""
    for field in UNWANTED_JSON_DIFF_FIELDS:
        if field in left:
            del left[field]
        if field in right:
            del right[field]

    return _diff(
        json.dumps(left, sort_keys=True, indent=4),
        json.dumps(right, sort_keys=True, indent=4),
        context,
    )


def difference(original: dict, revision: dict, context: int) -> dict:
    """Finds differences in a set of fields and the JSON itself"""
    diffs = {}

    for field in REVISION_FIELDS:
        if (field in original and field in revision) and (
            original[field] and revision[field]
        ):
            diff = _diff(original[field], revision[field])
            diffs[field] = diff

    diffs["json"] = _json_diff(original, revision)
    return diffs


def revision_before(latest_revision: int, revisions: list) -> dict:
    """Finds the revision before the current revision"""
    for revision in reversed(revisions):
        if revision["revision"] < latest_revision:
            return revision
    raise ValueError("No earlier revision found")


def parse_filter(filt: dict, context: int) -> Filter:
    """Parses a filter to a more human-readable format
    :param context:
    """
    enabled = filt["enabled"]
    updated = filt["revision"] != 1  # Revision 1 means this is the first revision
    diff = {}

    if updated and enabled:
        log.debug("Getting latest revision for filter %s" % filt["id"])
        revision = revision_before(
            filt["revision"], filter_api.revisions(id=filt["id"])["data"]
        )
        diff = difference(filt, revision, context)

    customer = filt["customer"]["shortName"] if filt["customer"] else "All customers"
    user = (
        filt["lastUpdatedByUser"]["userName"] if "lastUpdatedByUser" in filt else "N/A"
    )
    return Filter(
        filt["id"], customer, user, filt["description"], enabled, updated, diff
    )


def filter_filters(key: str, filters: list) -> list:
    """Filter away unwanted filters that are not in the given key

    The event-filter endpoint is old and doesn't provide a good way of doing this.
    """
    for filt in filters:
        if filt["type"] not in FILTER_TYPES[key]:
            filters.remove(filt)

    return filters


@register_command(extending=("eventfilter"), module=argus_cli_module)
def get_updated(
    start: time.date_or_relative,
    end: time.date_or_relative,
    types: FILTER_TYPES.keys(),
    context: int = 0,
    out: OUTPUT_METHODS.keys() = "jira",
):
    """Gets updated, new or deleted event-filters from Argus

    :param start: Start time of the search
    :param end: End time of the search
    :param types: The type of filter to search for
    :param context: Amount of lines of context to display in diff
    :param out: The desired output format
    """
    log.info("Fetching event-filters...")
    filters = filter_api.search(
        limit=0,
        startTimestamp=start,
        endTimestamp=end,
    )["data"]

    log.info("Filtering event_filters...")
    filters = filter_filters(types, filters)

    log.info("Parsing event-filters...")
    parsed = []
    for filt in filters:
        log.debug("Parsing filter with ID %d" % (filt["id"]))
        parsed.append(parse_filter(filt, context))

    log.info("")
    print(OUTPUT_METHODS[out](list(Filter._fields), parsed))



import sys

from argus_api.api.datastores.v1.descriptor import get_descriptor
from argus_api.api.datastores.v1.store.store import get_entries_from_store_simplified

from argus_cli.plugin import register_command
from argus_cli.utils.formatting import formatted_string
from argus_cli.helpers.pagination import offset_paginated
from argus_plugins import argus_cli_module

from argus_plugins.datastore import synchronize


@register_command(extending="datastores", module=argus_cli_module)
def dump(datastore, key: str = None, customer: str = None):
    """Displays a more human-readable version of a datastore

    :param datastore: The datastore to display
    :param list key: Filter to specific keys
    :param list customer: Customer(s) to filter by
    """
    store_info = get_descriptor(datastore)["data"]
    data = []
    pages = offset_paginated(get_entries_from_store_simplified)(
        dataStore=datastore, key=key, customerID=customer
    )

    for page in pages:
        data.extend(page["data"])

    print(
        "--- Datastore Metadata ---\n"
        "Datastore: {datastore}\n"
        "Type: {type}\n"
        "Behaviour: {behaviour}\n"
        "Global: {is_global}\n"
        "Lifetime: {lifetime}\n".format(
            datastore=datastore,
            type=store_info["dataType"],
            behaviour=store_info["behaviourType"],
            is_global=store_info["globalData"],
            lifetime=store_info["lifeTime"],
        )
    )

    # When the datastore is marked as "global data", no customer info is
    # attached. So in these cases we don't want to add the customer info to
    # the output.
    format_string = "{customer[name]} - " if not store_info["globalData"] else ""
    format_string += "{key}" if store_info["dataType"] == "LIST" else "{key}: {value}"

    print("----- Datastore Data -----\n" + formatted_string(data, format_string))


@register_command(extending="datastores", module=argus_cli_module)
def min_count(datastore: str, minimum_count: int):
    """Checks the size of a datastore.

    This command is typically used for tests in Nagios.
    It will exit with a failed state if the minimum size is less than the minimum

    :param datastore: The datastore to check
    :param minimum_count: The minimum allowed objects in the store.
    """
    # The limit is 1 because we don't care about the objects, only the count.
    count = get_entries_from_store_simplified(datastore, limit=1)["count"]

    if count >= minimum_count:
        print("OK - Found %d entries in datastore %s" % (count, datastore))
        sys.exit(0)
    else:
        print(
            "CRITICAL - Found %d entries in datastore %s (threshold=%s)"
            % (count, datastore, minimum_count)
        )
        sys.exit(2)



from typing import Callable, Iterable, Dict, Any, Optional, Type

from argus_cli.helpers.log import log


def retry(
    func: Callable,
    args: Optional[Iterable[Any]] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    exception_classes: Iterable[Type[Exception]] = None,
    max_retries: int = 0,
):
    """
    Retries calling a function up to a maximum number of attempts.

    :param func: function to retry
    :param args: positional arguments for the function
    :param kwargs: keyword arguments for the function
    :param exception_classes: iterable of exceptions to retry on.
    :param max_retries: maximum number of retries (mot including the initial call)
    :return: the return value of ``func(*args, **kwargs)``
    """
    args = args or tuple()
    kwargs = kwargs or dict()
    exception_classes = exception_classes or (Exception,)
    for try_ in range(max_retries + 1):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if (
                any(isinstance(e, exc) for exc in exception_classes)
                and try_ < max_retries
            ):
                log.warning(
                    f"retry: attempt({try_+1}/{max_retries}) failed, retrying. function={func}, args={args}, kwargs={kwargs}. Error: {e}."
                )
                continue
            raise



"""Utility functions for parsing and formatting time in plugins

If you're wondering why we're multiplying by 1000,
it's because Argus is written in Java and requires milis, not seconds.
"""
from argparse import ArgumentTypeError
from datetime import datetime, timedelta

import dateparser


def time_parser(datetime_string: str):
    """Turns a human readable time string into a datetime object.

    NOTE
    ====
    This is strictly a helper. If you want to have a normal datetime object
    in your code. Please just reference to a datetime object. Example:

    >>> def my_funky_func(start: datetime): pass

    All parsing will be done for you, so don't worry about that.

    :param datetime_string: DateTime string that will be parsed
    :return:
    """
    parsed = dateparser.parse(datetime_string)

    if parsed is None:
        raise ArgumentTypeError(
            "Could not parse provided time string."
            "\n\n"
            "A lot of different formats are supported. Both absolute and "
            "relative time strings. Here are some examples:\n"
            '"2 weeks ago", "1 minute ago", "2019-01-27", '
            '"1st of August 10:15 pm".'
            "\n\n"
            "If you have spaces in the time string, make sure that it is "
            'surrounded by a pair of double quotes ("like this")'
        )

    return parsed


def date_or_relative(datetime_string: str):
    """Convince function that merges date-time and time diff formats"""
    return int(time_parser(datetime_string).timestamp() * 1e3)


def time_diff(datetime_string: str):
    """Converts the input from one unit to milliseconds"""
    parsed_input = time_parser(datetime_string)
    return int((datetime.now() - parsed_input).total_seconds() * 1e3)


def timestamp_to_period(timestamp: int, format="iso"):
    """Converts a timestamp to a ISO8601 style period with days"""
    clock_time = datetime.utcfromtimestamp(timestamp)
    if format == "iso":
        return "P{days:03}DT{hours:02}:{minutes:02}:{seconds:02}".format(
            days=timedelta(seconds=timestamp).days,
            hours=clock_time.hour,
            minutes=clock_time.minute,
            seconds=clock_time.second,
        )
    if format == "minutes":
        return "{minutes} minutes".format(
            minutes=timedelta(seconds=timestamp).total_seconds() // 60,
        )


def timestamp_to_date(timestamp: int):
    """Converts a timestamp to a ISO8601 style date and time"""
    return datetime.fromtimestamp(timestamp).isoformat()


def datetime_to_week_number(dt: datetime) -> str:
    """returns the week number a datetime is in as 0-padded string."""
    return dt.strftime("%W")



from argus_cli.plugin import run, import_external_plugins

from argus_plugins import argus_cli_module
from argus_plugins.api_provider import import_submodules

# Import commands so that they get registered
from argus_plugins import *


def main():
    import_submodules("argus_api.api")
    import_external_plugins("argus_cli.plugins")
    run(argus_cli_module)


if __name__ == "__main__":
    main()



import json
import sys
from functools import reduce
from operator import getitem
from pprint import pprint

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from argus_cli.arguments import parse_arguments, Module, Provider, Command
from argus_cli.helpers.formatting import to_caterpillar_case
from argus_cli.helpers.log import log


def import_external_plugins(entry_point_name: str):
    """Import other plugins that are installed on the system.

    The implementation is done with reference to this python guide:
    https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata

    TODO: Investigate a better placement for loading these external plugins.
    """

    return {
        entry_point.name: entry_point.load()
        for entry_point in entry_points(group=entry_point_name)
    }


def run(func: callable = None, argv: list = None):
    """Will run the given command or module with commandline arguments

    Works with functions that have been registered with:
        register_command()
    Or without any arguments when a module has been initialized with:
        create_module()

    :param func: A function that has been registered.
    :raises ValueError: When the given function is not registered.
    """

    command, arguments = parse_arguments(func, argv)

    log.debug('Running command "{}"'.format(command))
    result = command(**arguments)

    if isinstance(result, dict):
        print(json.dumps(result, indent=2))
    elif result:
        pprint(result)


def create_module(providers: list = None) -> Module:
    """Set up for running a set of commands in a suite.

    Technical overview:
        Creates the commands module in the argument handler.
        This has to be created before any commands are registered.

    :param providers: All providers that this module requires to run.
    :param entry_point: The name of the entry_point of external plugins.
    """
    if providers is None:
        providers = []

    log.debug("Setting up module for commands")
    module = Module(providers=[provider._provider for provider in providers])

    return module


def register_command(
    alias: str = None,
    extending: tuple = None,
    module: Module = None,
    providers: list = None,
) -> callable:
    """Marks the given function as a command.

    Allows for calling the function on the commandline.
    Will also automatically add the module's providers to the command.

    :param alias: If the user wants a custom name on the plugin
    :param extending: A existing plugin to extend
    :param module: The base module to register against. If None it will register against the base module.
    :param providers: List of Providers that are required to call the function.
    """
    extending = (extending,) if isinstance(extending, str) else extending
    providers = providers or []
    base_module = module

    def decorate(func):
        # Rename all plugins and commands to caterpillar-case-format to conform to common cli naming.
        plugin_name = tuple(map(to_caterpillar_case, extending or (func.__module__,)))
        command_name = to_caterpillar_case(alias or func.__name__)

        log.debug(
            'Registering command "%s" in plugin "%s"'
            % (command_name, "/".join(plugin_name))
        )

        if base_module is not None:
            sub_module = reduce(getitem, plugin_name, base_module)

            command = Command(command_name, func, sub_module, providers=providers)
            sub_module[command_name] = command
        else:
            command = Command(command_name, func, providers=providers)

        # TODO: Should probably return the command in stead.
        func._command = command

        return func

    return decorate


def register_provider(name: str = None) -> callable:
    """Marks the given function as a provider.

    Providers allow extra functionality to be executed before another command.

    :param name: Alternative name for the module
    """

    def decorate(func):
        provider_name = to_caterpillar_case(name or func.__name__)

        log.debug('Registering provider "{}"'.format(name))
        func._provider = Provider(name, func)

        return func

    return decorate



import json
from datetime import datetime

from argus_api.api.customernetworks.v1.network import add_customer_network
from argus_api.exceptions.http import ArgusException
from argus_cli.helpers import formatting
from argus_cli.helpers.formatting import success, failure

from argus_cli.helpers.log import log
from argus_plugins.customer_networks.helpers.diff import diff_table


def create(
    networks: list,
    customer: dict,
    dry: bool = False,
    as_json: bool = False,
    skip_confirmation: bool = False,
) -> list:
    """Creates networks on Argus from a given list of CustomerNetworks

    :param list networks: List of CustomerNetworks
    :param bool as_json: Messages are returned as JSON
    :returns: Summary report as a list of messages
    """

    # Print updates and let user accept before performing updates
    summary = []

    # If we have networks to create, and we're not running in interactive mode,
    # perform the action on all networks without asking the user for each
    # network. Only ask once, after showing the difference.

    # Print the difference table showing which networks will be
    # updated and with what new values they will be updated
    print(
        formatting.table(
            title="New networks for %s" % customer["name"],
            keys=("IP", "Description", "Zone", "Location", "Flags"),
            data=diff_table(networks, show_only_changes=True),
        )
    )

    # Ask the user to confirm the change if we're not running with
    # `skip_confirmation`
    if skip_confirmation or formatting.ask("Accept changes?"):
        log.info("Creating networks for %s" % customer["name"])

        for network in networks:
            log.info(
                "Creating network %s for %s"
                % (network.to_json()["networkAddress"], customer["name"])
            )

            try:
                # Only perform API actions when we're not running
                # in dry mode
                if not dry:
                    # json.dumps can't dump dict_items, so we force it to a dict first
                    # and filter out flagsToDisable
                    network_json = network.to_json()
                    kwargs = {
                        field: value
                        for field, value in network_json.items()
                        if field not in ("flagsToDisable", "flagsToEnable")
                    }
                    # Verified is its own field
                    kwargs["flagsToSet"] = [
                        f
                        for f in network_json.get("flagsToEnable", [])
                        if f != "UNVERIFIED"
                    ]
                    kwargs["verified"] = "UNVERIFIED" not in network_json.get(
                        "flagsToEnable", []
                    )

                    log.debug(json.dumps(kwargs))
                    add_customer_network(**kwargs)

                # Add success message to summary report
                if as_json:
                    summary.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": "Created %s"
                            % (network.to_json()["networkAddress"]),
                            "name": network["description"],
                            "ip": network.to_json()["networkAddress"],
                            "body": network.to_json(),
                            "updated": False,
                            "created": True,
                            "failure": False,
                            "deleted": False,
                        }
                    )
                else:
                    summary.append(
                        success(
                            "Network %s created on Argus"
                            % (network.to_json()["networkAddress"])
                        )
                    )
                    log.info(summary[-1])

            # If we received a HTTP error, add the error message
            # to the summary report and log an error.
            except ArgusException as error:
                if as_json:
                    summary.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": "Failed to create network %s on Argus: %s"
                            % (network.to_json()["networkAddress"], error),
                            "name": network["description"],
                            "ip": network.to_json()["networkAddress"],
                            "body": network.to_json(),
                            "updated": False,
                            "created": True,
                            "failure": False,
                            "deleted": False,
                        }
                    )
                else:
                    summary.append(
                        failure(
                            "Failed to create network %s on Argus: %s"
                            % (network.to_json()["networkAddress"], error)
                        )
                    )
                    log.error(summary[-1])

    return summary



"""Used to import assets from CSV files"""
import csv
import json
import re
from datetime import datetime

from argus_api.api.assets.v1.host import (
    add_host_asset,
    update_host_asset,
    search_host_assets,
    delete_host_asset,
)
from argus_api.api.assets.v1.service import (
    search_service_assets,
    attach_hosts_to_service,
)
from argus_api.api.customers.v2.customer import get_customer

from argus_api.exceptions.http import ArgusException, ObjectNotFoundException
from argus_cli.helpers.collections import set_dotnotation
from argus_cli.helpers.formatting import ask, success, failure
from argus_cli.helpers.log import logging
from argus_cli.plugin import register_command
from argus_plugins import argus_cli_module

log = logging.getLogger("plugin").getChild("assets")


def diff(a: list, b: list):
    """Returns values from 'a' not found in 'b'"""
    return list(set(a) - set(b))


@register_command(extending=("assets", "import"), module=argus_cli_module)
def hosts(
    customer: str,
    file_path: str,
    map_headers: list = tuple(),
    extra_json: str = None,
    field_separator: str = " ",
    headers_on_first_line: bool = False,
    always_yes: bool = False,
    splunk: bool = False,
    replace_hosts: bool = False,
    output: str = None,
    dry: bool = False,
    delimiter: str = ",",
):
    """Imports assets from a CSV file

    Host assets *must* provide fields 'name' and 'ipAddresses'. If the file does not provide these fields, but have
    other names for these fields such as 'ip' and 'description', pass --map-headers ip:ipAddresses,description:name to
    map the headers to these names.

    A host may have multiple ipAddresses, these should be separated by a semicolon, like: 10.0.15.12;88.283.39.12

    Optional fields are:
        - operatingSystemCPE
        - type (SERVER or CLIENT, default: SERVER)
        - source (CVM or USER, default: USER)
        - aliases (list of aliases, separated by given field separator)
        - properties (additional properties to add to the host, formatted as JSON)

    You can add these fields using --extra-json type:CLIENT (to apply to all hosts), or as its
    own field in the CSV file.

    :param customer: Name of the customer to import assets for
    :param file_path: Path to the CSV file
    :param map_headers: Optional map of header names, e.g ipAddressv4:address,long_description:description
    :param field_separator: Separator used inside fields, e.g when providing multiple IP addresses or aliases. Defaults to whitespace.
    :param extra_json: Adds extra field: values to each JSON object read from CSV. Can be used to add missing fields
    :param headers_on_first_line: Whether headers should be taken from the first line in the CSV
    :param splunk: If this flag is set, JSON data will be written and the log will be suppressed
    :param replace_hosts: If this flag is set, if an IP belongs to another host, that host will be deleted before creating a new host
    :param output: File to write results to (used in conjunction with the splunk flag)
    :param dry: If this is enabled, no modifying API calls will be made
    :param delimiter: Delimiter in the csv. Default is "," but can be
    changed if the user would like.
    """

    headers = None

    # If headers are not on the first line of the CSV file
    # and we have a map of headers, create the header names
    # either by getting the last value of every old:new pair,
    # or just each value if there's no mapping
    if map_headers:
        headers = {
            (header.split(":")[0] if ":" in header else header): (
                header.split(":")[1] if ":" in header else header
            )
            for header in map_headers
        }

    with open(file_path) as assets_file:
        try:
            assets = [
                asset
                for asset in csv.DictReader(
                    assets_file,
                    delimiter=delimiter,
                    fieldnames=headers if not headers_on_first_line else None,
                )
            ]
        except UnicodeDecodeError:
            raise ValueError(
                "This file seems to be corrupt, it contains strange bytes."
                "Please check the file encoding, and try to re-save the file"
            )

    if not assets and headers_on_first_line and not map_headers:
        raise ValueError(
            "CSV file was empty, or contained only one row "
            "but you did not define any headers so these were "
            "thought to be headers"
        )

    # These fields are required
    for host in assets:
        if "name" not in host or "ipAddresses" not in host:
            raise ValueError(
                f"Required fields 'name' "
                f"and/or 'ipAddresses' not in {str(host)}"
                f"This might also be caused if you did not use the same "
                f"delimiter as specified or default (,) in the csv."
            )
    messages = []

    # Get the customer, and fail if none was found
    try:
        customer = get_customer(idOrShortName=customer)
    except ObjectNotFoundException:
        raise LookupError(f"No customer found for {customer}")

    # Search for all existing assets with the names we found
    existing_assets = search_host_assets(
        keywords=[host["name"] for host in assets if "name" in host],
        customerID=[customer["data"]["id"]],
        limit=10000,
    )

    # ... and create a lookup table
    existing_assets = {asset["name"]: asset for asset in existing_assets["data"]}

    # ... also create a lookup table for IPs to allow removing old hosts
    existing_ip_assets = {}

    for host in existing_assets.values():
        for ipAddress in host["ipAddresses"]:
            existing_ip_assets[ipAddress["address"]] = host

    # Update asset dicts with names from map_headers
    if map_headers:
        for host in assets:
            for key, new_key in headers.items():
                host[new_key] = host.pop(key)

    # Merge IP addresses of assets with the same name
    uniqueHosts = {host["name"]: host for host in assets if "name" in host}

    for host in assets:
        uniqueHosts[host["name"]]["ipAddresses"] = uniqueHosts[host["name"]][
            "ipAddresses"
        ].split(field_separator) + host["ipAddresses"].split(field_separator)
        uniqueHosts[host["name"]]["ipAddresses"] = field_separator.join(
            list(set(uniqueHosts[host["name"]]["ipAddresses"]))
        )

    services_to_attach_hosts = {
        # serviceName: [host1, host2]
    }

    for host in uniqueHosts.values():

        host.update(
            {
                "customerID": int(customer["data"]["id"]),
                "ipAddresses": host["ipAddresses"].split(field_separator),
                "properties": json.loads(host["properties"])
                if "properties" in host
                else {},
            }
        )

        if extra_json:
            host.update(json.loads(extra_json))

        # Replace keys with dot notation (e.g criticality.availability) with
        # a nested dict instead:
        for key in [key for key in host.keys() if "." in key]:
            host = set_dotnotation(dict(**host), key, host[key])
            del host[key]

        host["name"] = re.sub(
            r"[^\s\w\{\}\$-().\[\]\"\'_\/\\,\*\+#:@!\?;-]", "", host["name"]
        ).strip()

        # Check if any host already has any of these IPs, and ask the user if they want to remove
        # those hosts first
        for ip in host["ipAddresses"]:
            if ip in existing_ip_assets:
                existing_host = existing_ip_assets[ip]

                # Skip if the IPs are identical and the host name is identical
                if host["name"] == existing_host["name"] and sorted(
                    host["ipAddresses"]
                ) == sorted([ip["address"] for ip in existing_host["ipAddresses"]]):
                    continue

                # Keep track of host iD to associate services later on
                host.update({"id": existing_host["id"]})

                if (replace_hosts and always_yes) or ask(
                    f"A host already exists with IP {ip}"
                ):
                    # If the host has multiple IPs, update the host to remove this IP
                    if len(existing_host["ipAddresses"]) > 1:
                        try:
                            if not dry:
                                update_host_asset(
                                    existing_host["id"], deleteIpAddresses=[ip]
                                )
                            if splunk:
                                messages.append(
                                    {
                                        "timestamp": datetime.now().isoformat(),
                                        "message": f"Updated {existing_host['name']} and removed IP {ip}",
                                        "name": existing_host["name"],
                                        "ip": existing_host["ipAddresses"],
                                        "body": {
                                            "id": existing_host["id"],
                                            "deleteIpAddresses": [ip],
                                        },
                                        "updated": True,
                                        "deleted": False,
                                        "created": False,
                                        "failure": False,
                                    }
                                )
                            else:
                                log.plugin(
                                    success(
                                        f"Updated {existing_host['name']} and removed IP {ip}"
                                    )
                                )
                        except ArgusException as error:
                            if splunk:
                                messages.append(
                                    {
                                        "timestamp": datetime.now().isoformat(),
                                        "message": f"Failed to remove ip {ip} from {existing_host['name']}: {error}",
                                        "name": existing_host["name"],
                                        "ip": existing_host["ipAddresses"],
                                        "body": {
                                            "id": existing_host["id"],
                                            "deleteIpAddresses": [ip],
                                        },
                                        "updated": False,
                                        "created": False,
                                        "deleted": False,
                                        "failure": True,
                                    }
                                )
                            else:
                                log.plugin(
                                    failure(
                                        f"Failed to remove ip {ip} from {existing_host['name']}: {error}"
                                    )
                                )

                    # If the host only has one IP, delete the host altogether
                    else:
                        try:
                            if not dry:
                                delete_host_asset(existing_host["id"])
                            if splunk:
                                messages.append(
                                    {
                                        "timestamp": datetime.now().isoformat(),
                                        "message": f"Removed host {existing_host['name']}",
                                        "name": existing_host["name"],
                                        "ip": existing_host["ipAddresses"],
                                        "body": {
                                            "id": existing_host["id"],
                                        },
                                        "updated": False,
                                        "deleted": True,
                                        "created": False,
                                        "failure": False,
                                    }
                                )
                            else:
                                log.plugin(
                                    success(f"Removed host {existing_host['name']}")
                                )

                        except ArgusException as error:
                            if splunk:
                                messages.append(
                                    {
                                        "timestamp": datetime.now().isoformat(),
                                        "message": f"Failed to remove host {existing_host['name']}: {error}",
                                        "name": existing_host["name"],
                                        "ip": existing_host["ipAddresses"],
                                        "body": {
                                            "id": existing_host["id"],
                                        },
                                        "updated": False,
                                        "created": False,
                                        "deleted": False,
                                        "failure": True,
                                    }
                                )
                            else:
                                log.plugin(
                                    failure(
                                        f"Failed to remove ip {ip} from {existing_host['name']}: {error}"
                                    )
                                )

        # Add services to host if host has any services
        if "services" in host:
            for service in host["services"].split(";"):
                if service.strip() in services_to_attach_hosts:
                    services_to_attach_hosts[service.strip()].append(host)
                else:
                    services_to_attach_hosts[service.strip()] = [host]

        if host["name"] in existing_assets:
            host.update({"id": existing_assets[host["name"]]["id"]})

            if always_yes or ask(
                f"{host['name']} already exists, do you want to update it?"
            ):
                current_asset = existing_assets[host["name"]]
                try:
                    current_ips_with_range = [
                        f"{ip['address']}/{ip['maskBits']}"
                        for ip in current_asset["ipAddresses"]
                    ]
                    given_ips_with_range = [
                        f"{ip}/32" if "/" not in ip else ip
                        for ip in host["ipAddresses"]
                    ]

                    host.update(
                        {
                            "addIpAddresses": diff(
                                given_ips_with_range, current_ips_with_range
                            ),
                            "deleteIpAddresses": diff(
                                current_ips_with_range, given_ips_with_range
                            ),
                        }
                    )

                    if "aliases" in host:
                        host.update(
                            {
                                "addAliases": diff(
                                    host["aliases"].split(field_separator),
                                    [
                                        alias["fqdn"]
                                        for alias in current_asset["aliases"]
                                    ],
                                ),
                                "deleteAliases": diff(
                                    [
                                        alias["fqdn"]
                                        for alias in current_asset["aliases"]
                                    ],
                                    host["aliases"].split(field_separator),
                                ),
                            }
                        )

                    if "properties" in host:
                        host.update(
                            {
                                "addProperties": {
                                    property: value
                                    for property, value in host["properties"].items()
                                    if property
                                    in diff(
                                        host["properties"].keys(),
                                        current_asset["properties"].keys(),
                                    )
                                },
                                "deleteProperties": diff(
                                    current_asset["properties"].keys(),
                                    host["properties"].keys(),
                                ),
                            }
                        )

                    # Skip updating if nothing has changed
                    criticality_unchanged = True
                    if "criticality" in host:
                        try:
                            for key, value in existing_host["criticality"].items():
                                if key not in host["criticality"]:
                                    host["criticality"][key] = value
                                    criticality_unchanged = False
                                elif value != host["criticality"][key]:
                                    criticality_unchanged = False
                        except NameError:
                            # The host/existing_host has not been defined
                            criticality_unchanged = False

                    # Skip updating if nothing has changed
                    if (
                        not host["addIpAddresses"]
                        and not host["deleteIpAddresses"]
                        and not host.get("addAliases")
                        and not host.get("deleteAliases")
                        and not host.get("addProperties")
                        and not host.get("deleteProperties")
                        and host["description"] == current_asset["description"]
                        and criticality_unchanged
                    ):
                        continue

                    if not dry:
                        update_host_asset(
                            existing_assets[host["name"]]["id"],
                            **{
                                field: value
                                for field, value in host.items()
                                if field
                                not in (
                                    "ipAddresses",
                                    "customerID",
                                    "properties",
                                    "aliases",
                                    "services",
                                    "id",
                                )
                            },
                        )
                    if splunk:
                        messages.append(
                            {
                                "timestamp": datetime.now().isoformat(),
                                "message": f"Updated {host['name']}",
                                "name": host["name"],
                                "ip": host["ipAddresses"],
                                "body": {
                                    field: value
                                    for field, value in host.items()
                                    if field
                                    not in (
                                        "ipAddresses",
                                        "customerID",
                                        "properties",
                                        "aliases",
                                        "services",
                                        "id",
                                    )
                                },
                                "updated": True,
                                "created": False,
                                "deleted": False,
                                "failure": False,
                            }
                        )
                    else:
                        log.plugin(success(f"Updated {host['name']}"))

                except ArgusException as error:
                    if splunk:
                        messages.append(
                            {
                                "timestamp": datetime.now().isoformat(),
                                "message": f"Failed to update {host['name']}: {error}",
                                "name": host["name"],
                                "ip": host["ipAddresses"],
                                "body": {
                                    field: value
                                    for field, value in host.items()
                                    if field
                                    not in (
                                        "ipAddresses",
                                        "customerID",
                                        "properties",
                                        "aliases",
                                        "services",
                                        "id",
                                    )
                                },
                                "updated": False,
                                "created": False,
                                "deleted": False,
                                "failure": True,
                            }
                        )
                    else:
                        log.plugin(failure(f"Failed to update {host['name']}: {error}"))

        else:
            try:
                if not dry:
                    created_asset = add_host_asset(
                        **{
                            key: value
                            for key, value in host.items()
                            if key not in ("services", "id")
                        }
                    )

                    if "data" in created_asset and created_asset["data"]:
                        host["id"] = created_asset["data"]["id"]
                if splunk:
                    messages.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": f"Created {host['name']} ({', '.join(host['ipAddresses'])})",
                            "name": host["name"],
                            "ip": host["ipAddresses"],
                            "body": host,
                            "updated": False,
                            "created": True,
                            "failure": False,
                            "deleted": False,
                        }
                    )
                else:
                    log.plugin(
                        success(
                            f"Created {host['name']} ({', '.join(host['ipAddresses'])})"
                        )
                    )

            except ArgusException as error:
                if splunk:
                    messages.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": f"Failed to create {host['name']}: {error}",
                            "name": host["name"],
                            "ip": host["ipAddresses"],
                            "body": host,
                            "updated": False,
                            "created": False,
                            "failure": True,
                            "deleted": False,
                        }
                    )
                else:
                    log.plugin(failure(f"Failed to create {host['name']}: {error}"))

        if splunk and messages:
            if output:
                with open(output, "w") as json_file:
                    json_file.write(json.dumps(messages, indent=4, sort_keys=True))
            else:
                print(json.dumps(messages))

    for service_name, host_list in services_to_attach_hosts.items():
        try:
            for service in search_service_assets(
                name=[service_name], customerID=[int(customer["data"]["id"])]
            )["data"]:
                if service["name"] == service_name:
                    try:
                        if not dry:
                            attach_hosts_to_service(
                                id=service["id"],
                                hostAssetIDs=[
                                    host["id"] for host in host_list if "id" in host
                                ],
                            )
                            if splunk:
                                print(
                                    json.dumps(
                                        {
                                            "timestamp": datetime.now().isoformat(),
                                            "message": f"Attached {len(host_list)} hosts to service {service_name}",
                                            "name": service_name,
                                            "hosts": host_list,
                                            "updated": True,
                                            "created": False,
                                            "deleted": False,
                                            "failure": False,
                                        }
                                    )
                                )
                            else:
                                log.plugin(
                                    failure(
                                        f"Attached {len(host_list)} hosts to service {service_name}"
                                    )
                                )

                    except ArgusException as error:
                        if splunk:
                            print(
                                json.dumps(
                                    {
                                        "timestamp": datetime.now().isoformat(),
                                        "message": f"Failed to attach hosts to service {service_name}: {error}",
                                        "name": service_name,
                                        "hosts": host_list,
                                        "updated": False,
                                        "created": False,
                                        "deleted": False,
                                        "failure": True,
                                    }
                                )
                            )
                        else:
                            log.plugin(
                                failure(
                                    f"Failed to attach hosts to service {service_name}: {error}"
                                )
                            )

        except ArgusException as error:
            if splunk:
                print(
                    json.dumps(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": f"Failed to look up service {service_name}: {error}",
                            "name": service_name,
                            "updated": False,
                            "created": False,
                            "deleted": False,
                            "failure": True,
                        }
                    )
                )
            else:
                log.plugin(
                    failure(f"Failed to look up service {service_name}: {error}")
                )



"""Common formatting helpers for inputs and outputs to allow for a common interface across commands."""
import json
from functools import partial

from argus_cli.utils.formatting.data_formats import csv, jira_table, formatted_string

#: supported output formatters
FORMATS = {
    "csv": csv,
    "json": partial(json.dumps, indent=2),
    "jira-table": jira_table,
}


def get_data_formatter(format: str) -> callable:
    """Convenience function to get a format

    If the given format isn't known, it will be interpreted as a formatted string.

    :param format: A format string or one of the available formats
    """
    try:
        return FORMATS[format]
    except KeyError:
        return partial(formatted_string, format_string=format)



"""This is a helper module for all things case related"""

from typing import Union

from argus_api.api.customers.v2.customer import get_customer

#: Valid Argus case statuses
STATUSES = [
    "pendingCustomer",
    "pendingSoc",
    "pendingVendor",
    "workingSoc",
    "workingCustomer",
    "pendingClose",
    "closed",
]
#: Valid Argus case types
CASE_TYPES = ["securityIncident", "operationalIncident", "informational", "change"]
#: Valid Argus case priorities
PRIORITIES = ["low", "medium", "high", "critical"]
KEYWORD_FIELDS = ["subject", "description", "comments", "id", "all"]


def customer_from_shortname(name: str) -> dict:
    customer = get_customer(idOrShortName=name.lower())["data"]
    return customer


def get_customer_id(name: str) -> int:
    """Gets a customer's ID from their name

    :param name: The name of the customer
    """
    return customer_from_shortname(name)["id"]


def format_tags_list(tags: Union[list, dict], key_str="key"):
    """Converts a set of tags to the format expected by the API.

    Allows for a more lax format where the API expects a list looking like:

    ``[{"key": key, "value": value}, {"key": key, "value": value}]``

    This will allow:

    - ``[{key: value}, {"key": key, "value": value}]``
    - {key: value, key: value}

    :param tags: a list of tags in the "lax" format
    :param key_str: The string to use as dict keys in the result object.
    :return: a list of tags in the "strict" (API) format
    :raises ValueError: if the input is invalid
    """
    if isinstance(tags, dict):
        length = len(tags)
        if (length == 2 and all(k in (key_str, "value") for k in tags)) or length == 1:
            # we have either one short-form tag, or one long-form tag
            return [format_tag(tags, key_str=key_str)]
        # otherwise we have a set of short-form tags as a JSON object
        return [format_tag({k: v}, key_str=key_str) for k, v in tags.items()]

    elif isinstance(tags, list):
        return [format_tag(tag, key_str=key_str) for tag in tags]

    raise ValueError("Expecting a list or dictionary")


def format_tag(tag: dict, key_str: str = "key"):
    """Ensures that a case tag is properly formatted.

    The API expects tags in the following format:

    {"key": key, "value": value}

    This utility allows using the short {"key": value} form by converting it to the
    format expected by the API. If the tag already has that format, it is not changed.

    Will raise a ValueError if the tag is in an unexpected format.

    :param tag: The tag in either the {"key": value} or the {"key": key, "value": value} form
    :param key_str: The string to use as dict keys in the result object.
    :return: The tag in the {"key": key, "value": value} form
    :raises ValueError: if the input is invalid
    """
    if len(tag) == 1:
        key = next(iter(tag))
        value = tag[key]
        return {key_str: key, "value": value}
    elif len(tag) == 2 and "key" in tag and "value" in tag:
        return tag
    else:
        raise ValueError(
            "Invalid tag format, expected"
            ' either {key: value} or {"key": key, "value": value}'
        )



"""Helpers for extending argument input functionality"""


def organize_parameters(include: dict, exclude: dict) -> dict:
    """Organizes parameters into argus_api format include and exclude parameters

    :param include: Parameters to include in the search
    :param exclude: Parameters to exclude in the search
    :returns: A dict that can be used on a argus_api endpoint.
    """
    include = include or {}
    exclude = exclude or {}

    parameters = {"subCriteria": []}

    for key, value in include.items():
        parameters[key] = value
    for key, value in exclude.items():
        parameters["subCriteria"].append({"exclude": True, key: value})

    return parameters



from datetime import datetime

from argus_api.api.customernetworks.v1.network import update_customer_network
from argus_api.exceptions.http import ArgusException
from argus_cli.helpers import formatting
from argus_cli.helpers.formatting import success, failure

from argus_cli.helpers.log import log
from argus_plugins.customer_networks.helpers.diff import diff_table


def update(
    networks: list,
    customer: dict,
    as_json: bool = False,
    dry: bool = False,
    skip_confirmation: bool = False,
) -> list:
    """Updates existing networks

    :param list networks: List of CustomerNetworks
    :param bool dry: Whether any actual changes should be performed or not
    :param bool interactive: Ask the user to continue on each network?
    :return list[str]: Summary of messages produced
    """

    # Find all networks that have changed in the given list:
    dirty_networks = [n for n in networks if n.is_dirty()]

    # This will hold our messages (failure, success, etc), which we display at the end
    summary = []

    # If we have dirty networks, and we're not running in interactive mode,
    # perform the action on all networks without asking the user for each
    # network. Only ask once, after showing the difference.
    if any(dirty_networks):

        # Print the difference table showing which networks will be
        # updated and with what new values they will be updated
        print(
            formatting.table(
                data=diff_table(dirty_networks),
                keys=("IP", "Description", "Zone", "Location", "Flags"),
                title="Changes to %s networks" % customer["name"],
            )
        )

        # Ask the user to confirm the change if we're not running with
        # `skip_confirmation`
        if skip_confirmation or formatting.ask("Accept changes?"):
            log.info("Updating networks for %s" % customer["name"])

            for network in dirty_networks:
                log.info(
                    "Updating network %s for %s"
                    % (network.to_json()["networkAddress"], customer["name"])
                )

                try:
                    # Only perform the actual update action when
                    # not running in dry mode:
                    if not dry:

                        # Update the network with its dirty
                        # (modified) fields before creating JSON
                        # from it
                        network.update(network._dirty)
                        update_data = {
                            key: value
                            for key, value in network.to_json().items()
                            # Allowed fields for update:
                            if key
                            in (
                                "networkID",
                                "location",
                                "description",
                                "flagsToEnable",
                                "flagsToDisable",
                                "zone",
                            )
                        }
                        update_customer_network(**update_data)

                    # Add success message to summary report
                    if as_json:
                        summary.append(
                            {
                                "timestamp": datetime.now().isoformat(),
                                "message": "Network %s updated on Argus"
                                % (network.to_json()["networkAddress"]),
                                "name": network["description"],
                                "ip": network.to_json()["networkAddress"],
                                "body": network.to_json(),
                                "updated": True,
                                "created": False,
                                "failure": False,
                                "deleted": False,
                            }
                        )
                    else:
                        summary.append(
                            success(
                                "Network %s updated on Argus"
                                % (network.to_json()["networkAddress"])
                            )
                        )
                        log.debug(summary[-1])

                # If we received a HTTP error, add the error message
                # to the summary report and log an error.
                except ArgusException as error:
                    if as_json:
                        summary.append(
                            {
                                "timestamp": datetime.now().isoformat(),
                                "message": "Failed to update network %s on Argus: %s"
                                % (network.to_json()["networkAddress"], error),
                                "name": network["description"],
                                "ip": network.to_json()["networkAddress"],
                                "body": network.to_json(),
                                "updated": False,
                                "created": False,
                                "failure": False,
                                "deleted": True,
                            }
                        )
                    else:
                        summary.append(
                            failure(
                                "Failed to update network %s on Argus: %s"
                                % (network.to_json()["networkAddress"], error)
                            )
                        )
                        log.error(summary[-1])
    return summary



import sys

import argparse
from argparse import ArgumentParser
from collections.abc import MutableMapping
from collections import defaultdict
from functools import reduce, partial
from operator import getitem

from argus_cli.parser import parse_function
from argus_cli.helpers.log import log


class _ModuleDefaultDict(defaultdict):
    """A defaultdict that is specialized for use with the Module class

    The key has to be passed to the module.
    The key isn't passed to the class in a normal defaultdict.
    """

    def __missing__(self, key):
        self[key] = self.default_factory(name=key)
        return self[key]


class _ModuleAction(argparse._SubParsersAction):
    """Stores the name of the plugin to a attribute on the namespace

    Will add the name of the module to the _modules attribute of the namespace.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        super().__call__(parser, namespace, values, option_string)
        try:
            namespace._modules.insert(0, values[0])
        except AttributeError:
            namespace._modules = [values[0]]


class Module(MutableMapping):
    """Storage for argument parsers.

    This is basically a dict with argparse support.
    """

    def __init__(self, name=None, module=None, providers=None):
        self.providers = providers or []

        provider_parsers = [provider.parser for provider in self.providers]
        if module:
            self.argument_parser = module.add_parser(name, parents=provider_parsers)
        else:
            # If there are no parents, then this is the root module
            self.argument_parser = ArgumentParser(parents=provider_parsers)

        self.subparser = self.argument_parser.add_subparsers(action=_ModuleAction)

        # Make sure that sub-modules automatically get a parent
        self._dict = _ModuleDefaultDict(
            partial(Module, module=self.subparser, providers=providers)
        )

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        if key in self._dict and (
            not isinstance(value, Command)
            or value.function.__code__ != self._dict[key].function.__code__
        ):
            # Due to the fact that plugin.py loads every single file,
            # a command can be loaded twice. Thus, we'll have to check the code object as well.
            raise KeyError("A function with the name {} already exists.".format(key))

        self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)


class Provider(object):
    """A function that provides some functionality to commands."""

    called = False

    def __init__(self, name, func, parents=None):
        self.name = name
        self.function = func
        self.parser = ArgumentParser(add_help=False, parents=parents or [])

        _register_arguments(self.parser, parse_function(func))

    def __call__(self, *args, **kwargs):
        # Make sure that the provider is only called once.
        # TODO: Should look into how to solve this whole provider stuff in a better way.
        if self.called:
            return
        self.called = True

        return self.function(*args, **kwargs)


class Command(object):
    """A callable with an accompanying argument parser."""

    def __init__(
        self, name: str, func: callable, module: Module = None, providers: list = None
    ):
        self.name = name
        self.function = func
        self.providers = providers or []
        if module is not None:
            self.providers.extend(module.providers)

        argparse_initializer = (
            module.subparser.add_parser if module is not None else ArgumentParser
        )
        self.argument_parser = argparse_initializer(
            name, parents=[provider.parser for provider in self.providers]
        )

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


def _register_arguments(parser: ArgumentParser, metadata: dict):
    """Registers the arguments to the given argument parser.

    Used for lazy parsing of the commands.
    """
    parser.help = metadata.get("help")
    parser.description = metadata.get("description")

    for argument, options in metadata["arguments"].items():
        names = options.pop("names")

        if options.get("required") is False:
            prefixed_names = []
            for name in names:
                prefix = "-" if len(name) == 1 else "--"
                prefixed_names.append(prefix + name)
            names = prefixed_names
        elif len(names) > 1:
            log.warn(
                "%s is a positional argument, and can thus not have an alias. Ignoring aliases."
                % names[0]
            )
            names = [names[0]]
        parser.add_argument(*names, **options)


def _error_or_help(parser):
    if any(keyword in sys.argv for keyword in ("--help", "-h")):
        parser.print_help()
        parser.exit()
    parser.error("Not enough arguments")


def _get_command_from_arguments(argv: list, module: Module):
    """Gets the command to execute based on arguments.

    This part skips the help-message, as we haven't parsed the command yet.
    """
    base_args = [arg for arg in argv if (arg != "--help" and arg != "-h")]
    _, base_args = parse_providers(base_args, module.providers)

    module_arguments, remaining_args = module.argument_parser.parse_known_args(
        base_args
    )
    if not hasattr(module_arguments, "_modules"):
        _error_or_help(module.argument_parser)

    command = reduce(getitem, module_arguments._modules, module)
    if not isinstance(command, Command):
        _error_or_help(command.argument_parser)

    # Because we're passing the remaining args around, we'll have to reintroduce the help message.
    if "--help" in argv or "-h" in argv:
        remaining_args.append("--help")

    return command, remaining_args


def parse_providers(argv: list, providers: list):
    """Parses and runs the given providers

    :returns: Remaining arguments.
    """
    provider_arguments = {}
    for provider in providers:
        args, argv = provider.parser.parse_known_args(argv)
        provider_arguments.update(vars(args))
        provider(**vars(args))

    return provider_arguments, argv


def _parse_command(argv: list, command: Command):
    """Gets arguments that can be passed to the callable command

    Also runs the providers that the command requires.

    NOTE:
        That we're running the providers from here might prove confusing.
        Should be looked into at some point.
    """
    provider_arguments, argv = parse_providers(argv, command.providers)
    arguments = vars(command.argument_parser.parse_args(argv))

    for arg in provider_arguments:
        del arguments[arg]

    return arguments


def parse_arguments(func: [Module, Command], argv: list = None) -> (callable, dict):
    """Two parts: Module args and command args

    :param func: A function with a
    :param argv: Arguments to parse. If none, sys.argv will be used.
    """
    # This can't be set as a default in the signature.
    # Seems like it isn't initialized when the module gets loaded.
    argv = sys.argv[1:] if argv is None else argv

    if hasattr(func, "_command"):
        func = func._command
        first_argument_parse = lambda argv, func: (func, argv)
    elif isinstance(func, Module):
        first_argument_parse = _get_command_from_arguments
    else:
        raise ValueError(
            "The first argument has to be a Module or a Command. Given {}".format(
                type(func)
            )
        )

    # Grabs the correct command and removes the module arguments.
    command, remaining_args = first_argument_parse(argv, func)

    # Registering the metadata is done at a later point to not do unnecessary operations with _parse_function().
    # A single call to _parse_function() takes quite some time, so it's better to do it JIT.
    log.debug("Registering metadata for {}".format(command))
    _register_arguments(command.argument_parser, parse_function(command.function))

    log.debug("Parsing command arguments")
    arguments = _parse_command(remaining_args, command)

    return command, arguments



from string import Template

AUTOCREATE_OUTPUT_TEMPLATE = Template(
    """
$subject
$subject_line

Description
$description_line
$description


Associated events 
$associated_events_line
$associated_events
"""
)



import argparse
import collections
import inspect
import re
import sys
from datetime import datetime
from functools import wraps, partial
from pydoc import locate

import yaml

from argus_cli.helpers.formatting import to_caterpillar_case
from argus_cli.helpers.log import log
from argus_cli.utils.time import time_parser


def parse_function(function: callable) -> dict:
    """Parses a functions parameters and help-text from its docstring and annotations.

    :param function: A function
    :return: The description and arguments (in order) for the function
    """
    parsed = {}

    log.debug("Parsing arguments and docstring for %s..." % function.__name__)

    parsed["arguments"] = _parse_parameters(function)

    if function.__doc__:
        parsed = _parse_docstring(function, parsed)

    return parsed


def _parse_parameter(name: str, parameter: inspect.Parameter) -> dict:
    """Parses a parameter"""
    log.debug("Parsing parameter %s" % name)

    options = {}
    if parameter.kind is parameter.VAR_POSITIONAL:
        # This is necessary due to the nargs argument would always be
        # overwritten if the parameter wasn't a vararg.
        options["number_of_arguments"] = "*"
    if parameter.default is not parameter.empty:
        # Same reasoning as above, would overwrite when it shouldn't
        options["default"] = parameter.default

    return get_argument_creator(
        # An empty annotation is "empty" in stead of None...
        parameter.annotation
        if parameter.annotation is not inspect.Parameter.empty
        else None
    )(
        name=name,
        required=parameter.default is parameter.empty
        or parameter.kind is inspect.Parameter.VAR_POSITIONAL,
        **options
    )


def _parse_parameters(function: callable) -> dict:
    """Parses a functions parameters.

    :param function: The function to parse
    :returns: All arguments, ready to pass to argparse
    """
    arguments = collections.OrderedDict()
    signature = inspect.signature(function)

    log.debug("%s arguments: %s" % (function.__name__, signature))

    for name, parameter in signature.parameters.items():
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            # **kwargs Have to be defined in docstrings
            log.debug(
                "**%s argument ignored. kwargs are added from the docstring." % name
            )
            continue
        arguments[name] = _parse_parameter(name, parameter)

    log.debug(
        "%s: Registered commands from signature:\n\t%s" % (function.__name__, arguments)
    )
    return arguments


def _parse_docstring(function: callable, parsed: dict) -> dict:
    """Parses a function's docstring for more info about it and it's parameters.

    :param function: The function to parse
    :param parsed: Existing arguments for the function
    :return: Short description, Long description and more argument info
    """

    def _parse_docstring_aliases(docstring: str) -> dict:
        """Parses the docstring for parameter aliases

        :param docstring: The docstring to parse
        """
        #: Gets aliases for a parameter
        alias_regex = re.compile(r":alias (?P<name>\w+):\s+(?P<aliases>.*)")

        for name, aliases in alias_regex.findall(docstring):
            if name not in parsed["arguments"]:
                raise NameError(
                    "%s is not an argument. An argument has to exist to be aliased."
                    % name
                )
            for alias in aliases.split(","):
                parsed["arguments"][name]["names"].append(
                    to_caterpillar_case(alias.strip())
                )

        return parsed["arguments"]

    def _parse_docstring_parameters(docstring: str) -> dict:
        """Parses the docstring for extra info about parameters.

        :param docstring: The docstring to parse
        """
        #: Gets doc for a parameter. `argument_type` is optional.
        param_regex = re.compile(
            r":param\s?(?P<argument_type>\w*) (?P<name>\w+): (?P<doc>.*)"
        )

        for argument_type, name, doc in param_regex.findall(docstring):
            if argument_type and parsed["arguments"].get("type"):
                log.debug(
                    "Argument %s's type is set in both function annotation and docstring. Prioritizing docstring."
                    % name
                )

            new_argument = get_argument_creator(
                locate(argument_type) if argument_type else None, False
            )(
                parsed["arguments"].get(name, {}).get("names", [name])[0],
                required=None if parsed["arguments"].get(name) else False,
                action=parsed["arguments"].get(name, {}).get("action"),
            )
            argument = parsed["arguments"].setdefault(name, new_argument)

            if argument is not new_argument:
                argument.update(new_argument)
            if "type" not in argument and argument_type and "action" not in argument:
                # Special case to handle when there is no type from before
                argument["type"] = _str_or_file

            argument["help"] = doc

        return parsed["arguments"]

    # Escape all % { and } so argparse doesnt crash when trying to format the string
    docstring = (
        function.__doc__.replace("{", "{{")
        .replace("}", "}}")
        .replace("%", "%%")
        .split("\n", 1)
    )

    parsed["help"] = docstring[0].strip()

    if len(docstring) <= 1:
        return parsed

    description = docstring[1]

    match = re.compile(r":(?:param|alias)").search(description)
    arguments_part = description[match.start() :] if match else None
    parsed["description"] = (
        description[: match.start()] if match else description
    ).strip()

    if arguments_part:
        parsed["arguments"] = _parse_docstring_parameters(arguments_part)
        parsed["arguments"] = _parse_docstring_aliases(arguments_part)

    return parsed


def _create_argument(
    name: str,
    required: bool = None,
    type: callable = None,
    default: object = None,
    number_of_arguments: str = None,
    action: str = None,
    choices: list = None,
) -> dict:
    nick_name = to_caterpillar_case(name) if required is False else name
    if default is True:
        nick_name = "no-%s" % nick_name if len(nick_name) > 1 else nick_name

    argument = {"names": [nick_name]}

    if required is False:
        # Required can not be set for positionals, so we only set it if it's not required.
        argument["dest"] = name
        argument["required"] = False

    if default is not None:
        argument["default"] = default
    if number_of_arguments is not None:
        argument["nargs"] = number_of_arguments
    if action is not None:
        argument["action"] = action
    if choices is not None:
        argument["choices"] = choices
    if type is not None:
        argument["type"] = type

    return argument


@wraps(_create_argument)
def _handle_boolean(*args, **metadata):
    if metadata.get("default") is True:
        metadata["action"] = "store_false"
    elif metadata.get("default") is False:
        metadata["action"] = "store_true"
    elif "default" in metadata and metadata.get("default") is None:
        metadata["choices"] = [True, False]
        metadata["action"] = "store"
        return _create_argument(*args, type=_str_to_bool, **metadata)
    return _create_argument(
        *args, type=bool if not metadata.get("action") else None, **metadata
    )


def get_argument_creator(
    argument_type: type, assume_list_type: bool = True
) -> _create_argument:
    """Gets the correct create argument handler.

    This is necessary because some types needs special handling.

    :param argument_type: The type of the argument
    :param assume_list_type: If True, a list will automatically be assumed to be a string
    :returns: Argument creator function
    """
    if argument_type is sys.stdin:
        # This is a special case for stdin, where it's allowed to have specified parameter.
        # Normally the framework will parse instances of FileType just fine and accept stdin if the user uses "-".
        # See the argparse documentation on FileType.
        return partial(
            _create_argument,
            type=argparse.FileType("r"),
            default=sys.stdin,
            number_of_arguments="?",
        )
    elif inspect.isclass(argument_type) and issubclass(argument_type, datetime):
        return partial(_create_argument, type=time_parser)
    elif inspect.isclass(argument_type) and issubclass(argument_type, (list, tuple)):
        return partial(
            _create_argument,
            number_of_arguments="*",
            type=_str_or_file if assume_list_type else None,
        )
    elif inspect.isclass(argument_type) and issubclass(argument_type, dict):
        return partial(_create_argument, type=yaml.safe_load)
    elif isinstance(argument_type, collections.abc.Container) and not isinstance(
        argument_type, str
    ):
        if any(
            isinstance(element, collections.abc.Container)
            and not isinstance(element, str)
            for element in argument_type
        ):
            raise ValueError("A list of choices can not have a nested iterable object.")
        return partial(_create_argument, choices=argument_type)
    elif isinstance(argument_type, bool) or argument_type is bool:
        return _handle_boolean
    elif argument_type is str:
        return partial(_create_argument, type=_str_or_file)
    elif argument_type is not None and not callable(argument_type):
        raise ValueError(
            "Argument with type {} is not a valid argument type.".format(argument_type)
        )
    else:
        return partial(_create_argument, type=argument_type)


def _str_or_file(argument):
    """Helper type to parse a string as a file.

    This is because of the implementation of @<file> didn't work as expected originally,
    and this might not be the best practice way to do a cli application.
    """
    if argument.startswith("@") and argument != "@current":
        # @current is a special case (by convention, current API user)
        with open(argument[1:], "r") as f:
            output = "".join(f.readlines())
    else:
        output = argument

    return output


def _str_to_bool(argument):
    if isinstance(argument, bool):
        return argument
    if argument.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif argument.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError(
            "Boolean value expected. "
            "yes|true|t|y|1 for True, "
            "no|false|f|n|0 for False"
        )



from os.path import isfile
from typing import Union, List
from csv import DictReader
from argus_api.api.sensors.v1.location import get_location
from argus_api.exceptions.http import ArgusException
import functools


@functools.lru_cache(maxsize=0)
def resolve_location(customer, loc):
    if not customer:
        raise ValueError("expected customer")
    try:
        location = get_location(loc)["data"]
        if location and (
            location["customerInfo"]
            and location["customerInfo"]["id"] == customer["id"]
        ):
            return location["shortName"]
    except (ArgusException, IndexError):
        pass
    return "unknown"


class CustomerNetwork(dict):
    """A single customer network"""

    FLAG_KEYWORDS = {
        "CLIENT": ["CLIENT", "KLIENT"],
        "PROXY": ["PROXY"],
        "NAT": ["NAT", "HIDE"],
        "SCANNER": ["SKANNER", "SCANNER"],
        "DNS": ["DNS", "NAVNESERVER", "NAMESERVER"],
        "GUEST": ["GUEST", "GJEST"],
        "UNASSIGNED": ["UNASSIGNED"],
        "DELETED": ["DELETED", "FJERNET", "SLETTET", "REMOVED"],
    }

    def __init__(
        self,
        network_address: str,
        description: str,
        flags: (list, set) = (),
        customer: (str, None) = None,
        zone: str = "UNKNOWN",
        subnet_mask: (str, None) = None,
        **kwargs,
    ):
        super(CustomerNetwork, self).__init__()
        # remove empty strings from the flags list as they cause errors.
        flags = [f for f in flags if len(f)]
        self._dirty = {}
        self._fresh = True
        self.merge(
            {
                "location": {"shortName": kwargs.get("location", "unknown")},
                "zone": zone,
                "network_address": network_address.strip(),
                "description": description.strip(),
                "flags": flags,
                "customer": customer,
                "subnet_mask": subnet_mask,
            },
            **kwargs,
        )

        if self["maskBits"] and not isinstance(self["maskBits"], int):
            self["maskBits"] = int(self["maskBits"])

    def to_json(self) -> dict:
        """
        Converts CustomerNetwork to a JSON object acceptable by Argus

        :return: dict
        """
        _json = dict(**self)
        _json.update(self._dirty)

        json = {
            key: value
            for key, value in _json.items()
            if key
            in (
                "customer",
                "flagsToEnable",
                "flagsToDisable",
                "description",
                "location",
                "zone",
                "id",
            )
            # No need to provide flagsToEnable and flagsToDisable if these are empty:
            and not ("flagsTo" in key and not len(value))
        }
        json["networkAddress"] = "%s/%d" % (self["address"], int(self["maskBits"]))

        if "id" in json:
            json["networkID"] = json["id"]

        if "location" not in json:
            json["location"] = "Unknown"
        if "zone" not in json:
            json["zone"] = "UNKNOWN"

        return json

    def merge(self, customer_network: dict, **kwargs) -> object:
        """
        Updates this object with new values

        If any values existed before, and changed after the update, this CustomerNetwork is considered 'dirty'.

        :param dict customer_network: CustomerNetwork structure either as a dict, or as a CustomerNetwork object
        :return:
        """
        # If we have data set already in self, assign this data to _dirty instead, so that
        # we can keep track of which fields changed and only update those fields
        _attributes = self if self._fresh and len(self.items()) < 1 else self._dirty

        # If the other object has an ID attribute, reverse this merge instead, and merge this object into
        # the target object instead - that way, we maintain Argus' version as the source of truth
        if isinstance(customer_network, CustomerNetwork) and customer_network.exists():
            return customer_network.merge(self)
        elif (
            isinstance(customer_network, dict)
            and "id" in customer_network
            and customer_network["id"]
        ):
            return CustomerNetwork(customer_network).merge(self)

        # Normalize the input
        if isinstance(customer_network, CustomerNetwork):
            network_address = customer_network.to_json()["networkAddress"]
            description = customer_network["description"]
            flags = customer_network["flags"]
            customer = (
                customer_network["customer"] if "customer" in customer_network else None
            )
            subnet_mask = None
            zone = customer_network["zone"]
            location = customer_network["location"]
        else:
            network_address = customer_network["network_address"]
            description = customer_network["description"]
            flags = customer_network["flags"]
            customer = customer_network["customer"]
            subnet_mask = customer_network["subnet_mask"]
            zone = customer_network["zone"]
            location = customer_network["location"]["shortName"]

        _attributes["zone"] = (
            zone if zone in ("INTERNAL", "EXTERNAL", "DMZ") else "UNKNOWN"
        )
        _attributes["location"] = resolve_location(customer, location)

        # Check that network address is valid, and parse out /24 subnet bits from it if set
        if "/" in network_address:
            _attributes["address"], _attributes["maskBits"] = network_address.split("/")
        else:
            _attributes["address"] = network_address

        # Set description to _dirty
        _attributes["description"] = description

        # If subnet_mask is provided, check how many bits are set
        if subnet_mask:
            _attributes["maskBits"] = self._subnet_mask_to_bits(subnet_mask)
        elif "maskBits" not in _attributes:
            _attributes["maskBits"] = "32"

        # Extract possible flags from description
        flags_from_description = self._flags_from_description(description)

        # If we received flags, and this CustomerNetwork is a fresh one, just assign flags
        if self._fresh:
            _attributes["flags"] = flags or []
            _flags = set(flags + flags_from_description)

            # Parse flags from description and intersect them with the set flags
            self._dirty.update(self._diff_flags(_flags))

        # If this is not a fresh object, just update _attributes with the parsed flags
        else:
            _flags = set(flags + flags_from_description)
            _attributes.update(self._diff_flags(_flags))

        if customer and isinstance(customer, dict):
            _attributes["customer"] = customer["shortName"]
            _attributes["customerID"] = customer["id"]
        elif customer and isinstance(customer, str):
            _attributes["customer"] = customer

        for arg, value in kwargs.items():
            _attributes[arg] = value

        # Remove values in _dirty that are identical to self
        self._dirty = {
            k: v for k, v in self._dirty.items() if k not in self or self[k] != v
        }

        # Set _fresh to False if _attributes is _dirty, which it is if _fresh was True
        self._fresh = _attributes is self
        _attributes["flags"] = list(self["flags"])

        if "id" in kwargs:
            _attributes["id"] = int(kwargs["id"])

        return self

    def _diff_flags(self, flags: dict) -> dict:
        """
        Receives a set of flags, and compares them against the current flags, and returns a dict with flagsToEnable
        and flagsToDisable based on the received flags intersected with the existing flags

        :param set|list flags: Set or List of flags
        :return: dict
        """
        _attributes = {}

        # Flags enabled on the server, that are not enabled on file, will be disabled
        _attributes["flagsToDisable"] = set(self["flags"] or []).difference(flags or ())

        # Flags enabled in file, but not on server, will be enabled
        _attributes["flagsToEnable"] = set(flags or []).difference(self["flags"] or ())

        # Flags existing on both, will be kept
        _attributes["flagsToEnable"] = _attributes["flagsToEnable"] | set(
            self["flags"] or ()
        ).intersection(set(flags))

        # Cast back to list, because Set is not JSON serializable
        _attributes["flagsToEnable"] = list(_attributes["flagsToEnable"])
        _attributes["flagsToDisable"] = list(_attributes["flagsToDisable"])

        return _attributes

    def expected_flags(self) -> list:
        """
        Returns flags this CustomerNetwork will have after updating with flagsToEnable and flagsToDisable

        :return:
        """
        return list(
            set(
                set(self["flags"])
                | set(
                    self._dirty["flagsToEnable"]
                    if "flagsToEnable" in self._dirty
                    else []
                )
            )
            - set(
                self._dirty["flagsToDisable"] if "flagsToDisable" in self._dirty else []
            )
        )

    def exists(self) -> bool:
        """
        Checks if we have an ID on Argus, in which case this network currently exists

        :return:
        """
        return "id" in self and int(self["id"]) > 0

    def is_dirty(self) -> bool:
        """
        Checks if this network has been modified and needs to be updated on server, e.g any fields changed, or
        missing ID field

        :return:
        """
        # Compare everything in _dirty - if we're comparing flags/flagsToEnable/flagsToDisable, compare expected
        # flags after update with current flags by merging flagsToEnable / flagsToDisable into flags
        if "id" not in self or any(
            list(
                (
                    self["flags"] != self.expected_flags()
                    if "flags" in key
                    else value != self[key]
                    for key, value in self._dirty.items()
                    if key in self
                )
            )
        ):
            return True

        return False

    @staticmethod
    def from_json(
        json, customer=None
    ) -> Union["CustomerNetwork", List["CustomerNetwork"]]:
        """
        Converts JSON to CustomerNetwork. If json is a list, it converts all items and returns a list. Otherwise
        converts single item and returns CustomerNetwork

        :param dict[list json: Single object from server
        :return: list<CustomerNetwork>|CustomerNetwork
        """
        if isinstance(json, dict):
            return CustomerNetwork(
                network_address="%s/%d"
                % (
                    json["networkAddress"]["address"],
                    int(json["networkAddress"]["maskBits"]),
                ),
                location=json["location"]["shortName"],
                flags=json["flags"],
                description=json["description"],
                zone=json["zone"],
                id=json["id"],
                customer=customer,
            )
        elif isinstance(json, list):
            return [CustomerNetwork.from_json(single_object) for single_object in json]
        else:
            return []

    @staticmethod
    def from_csv(
        file_path: str,
        header_format: iter = ("address", "description"),
        headers_on_first_line: bool = False,
        delimiter: str = ",",
        customer: dict = {},
    ) -> list:
        """
        Parses a CSV file and returns a list of networks to be updated/created/replaced to Argus

        :param str file_path: Path to CSV file
        :param str delimiter: Delimiting character in CSV (default: ,)
        :param iterable header_format: List or tuple of headers for CSV file
        :param bool headers_on_first_line: Whether headers are on first line in CSV or not - if they're not, default headers are used
        :return: list[CustomerNetwork]
        """

        # Raise FileError if file does not exist
        if not isfile(file_path):
            raise FileNotFoundError("%s does not exist" % file_path)

        # Parse CSV file to list<network> dict format
        with open(file_path) as network_csv:
            reader = DictReader(
                network_csv,
                fieldnames=None if headers_on_first_line else header_format,
                delimiter=delimiter,
            )

            reader = list(reader)

            return [
                CustomerNetwork(
                    network_address=row.get("address") or row.get("Network"),
                    description=row.get("description") or row.get("Description"),
                    subnet_mask=row["subnet_mask"] if "subnet_mask" in row else None,
                    customer=customer,
                    zone=row.get("zone") or row.get("Zone", "UNKNOWN"),
                    location=row.get("LocationShortName")
                    or row.get("location", "unknown"),
                    flags=row.get("Flags", "").split("|"),
                )
                for row in reader
            ]

    # Private
    def _flags_from_description(self, description: str = "") -> list:
        """
        Parses a network string and sets flags depending on keywords found in string

        :param str network_string: String to parse
        :param dict flags: dict with { flag: [keyword1, keyword2] }
        :return: list
        """
        return [
            flag
            for flag, keyword_list in CustomerNetwork.FLAG_KEYWORDS.items()
            if any(
                (
                    keyword.lower() in (description or "").lower()
                    for keyword in keyword_list
                )
            )
        ]

    def _subnet_mask_to_bits(self, subnet_mask: str) -> int:
        """
        Returns the /bits for a subnet mask

        :param str subnet_mask:
        :return int:
        """
        return sum(map(lambda x: bin(int(x)).count("1"), subnet_mask.split(".")))

    def __contains__(self, other: ("CustomerNetwork", dict)) -> bool:
        """
        Checks if this CustomerNetwork contains the same network / bitmask, and, in effect, is a (potentially dirty)
        version of the other

        :param str|CustomerNetwork|dict other:
        :return:
        """
        if isinstance(other, str):
            return super(CustomerNetwork, self).__contains__(other)

        subject = other

        if "address" not in subject:
            return False

        if "/" in subject["address"]:
            subject["address"], subject["maskBits"] = subject["address"].split("/")

        if "subnet_mask" in subject:
            subject["maskBits"] = self._subnet_mask_to_bits(subject["subnet_mask"])

        return (
            "maskBits" in subject
            and subject["address"] == self["address"]
            and subject["maskBits"] == self["maskBits"]
        )

    def __eq__(self, other: ("CustomerNetwork", dict)) -> bool:
        """
        Compare this CustomerNetwork to another CustomerNetwork

        :param dict|CustomerNetwork other:
        :return:
        """
        other_attrs = {
            key: value
            for key, value in other.items()
            if key not in ("id", "customer", "customerID")
        }
        self_attrs = {
            key: value
            for key, value in self.items()
            if key not in ("id", "customer", "customerID")
        }
        return other_attrs == self_attrs

    def __ne__(self, other: ("CustomerNetwork", dict)) -> bool:
        """
        Compare this CustomerNetwork to another CustomerNetwork

        :param dict|CustomerNetwork other:
        :return:
        """
        return not self.__eq__(other)



from datetime import datetime

from argus_api.api.customernetworks.v1.network import delete_customer_network
from argus_api.exceptions.http import ArgusException

from argus_cli.helpers import formatting
from argus_cli.helpers.formatting import success, failure
from argus_cli.helpers.log import log


def destroy(
    networks: list,
    customer: dict,
    as_json: bool = False,
    dry: bool = False,
    skip_confirmation: bool = False,
) -> list:
    """Destroys existing networks

    :param list networks: List of CustomerNetworks
    :param dict customer: Customer object (used only for logging)
    :param bool as_json: Messages are returned as JSON
    :param callable authenticate: Authentication method to pass in authentication headers to API functions
    :param bool interactive: Whether or not to ask the user for each network. Defaults to asking once only.
    :returns: List of messages for summary report
    """
    # Print updates and let user accept before performing updates
    summary = []

    for index, network in enumerate(networks):
        # Clear terminal when interactive:
        formatting.clear()

        # Print table of networks that will be destroyed
        print(
            formatting.table(
                data=[network],
                keys=("IP", "Description", "Zone", "Location", "Flags"),
                title="Remove network %s (%d of %d)"
                % (network.to_json()["networkAddress"], index + 1, len(networks)),
                format=formatting.red,
            )
        )

        # Ask user to confirm the change unless we're running with alwaysYes
        if skip_confirmation or formatting.ask("Remove this network?"):
            log.info("Removing network for %s" % customer["name"])

            try:
                # Only perform the API action if we're not running in
                # dry mode, in which case no actual changes will be made
                if not dry:
                    delete_customer_network(id=network["id"])

                # Add a success message to the summary report
                # Add success message to summary report
                if as_json:
                    summary.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": "Removed %s"
                            % (network.to_json()["networkAddress"]),
                            "name": network["description"],
                            "ip": network.to_json()["networkAddress"],
                            "body": network.to_json(),
                            "updated": False,
                            "created": False,
                            "failure": False,
                            "deleted": True,
                        }
                    )
                else:
                    summary.append(
                        success(
                            "Network %s removed from Argus"
                            % (network.to_json()["networkAddress"])
                        )
                    )
                    log.info(summary[-1])

            # Catch any HTTP errors and add the error messages to the
            # summary report instead of crashing
            except ArgusException as error:
                if as_json:
                    summary.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": "Failed to remove network %s from Argus: %s"
                            % (network.to_json()["networkAddress"], error),
                            "name": network["description"],
                            "ip": network.to_json()["networkAddress"],
                            "body": network.to_json(),
                            "updated": False,
                            "created": False,
                            "failure": False,
                            "deleted": True,
                        }
                    )
                else:
                    summary.append(
                        failure(
                            "Failed to remove network %s from Argus: %s"
                            % (network.to_json()["networkAddress"], error)
                        )
                    )
                    log.error(summary[-1])
    return summary



import smtplib
import sys
import time
import json
from email.mime.text import MIMEText
from datetime import datetime, timedelta
from urllib.parse import quote

from argus_cli.plugin import run
from argus_cli.plugin import register_command
from argus_cli.utils.time import date_or_relative
from argus_cli.helpers.log import log
from argus_api.api.cases.v2.case import advanced_case_search
from argus_plugins import argus_cli_module

from argus_plugins.cases import utils
from argus_plugins.cases.utils import get_customer_id

UNASSIGNED = "unassigned"
SKIPABLES = ["tech", "assigned", "unassigned"]

DEFAULT_STATUS = utils.STATUSES[:-1]  # Exclude closed status


def parse_data(cases: list, skip_notify_tech: bool) -> dict:
    """Parses cases and associates them with a user

    :param cases: Cases to parse
    :param skip_notify_tech: If true, sort all assigned cases under "unassigned"
    :returns: All cases sorted after the user to send to (or "unassigned" to send to notify list)
    """
    parsed_cases = {UNASSIGNED: []}

    for case in cases:
        if case["assignedTech"] and not skip_notify_tech:
            tech = case["assignedTech"]["userName"]
            log.info("%s: Adding case #%s" % (tech, case["id"]))
            if tech not in parsed_cases:
                parsed_cases[tech] = []
            parsed_cases[tech].append(case)
        else:
            log.info("%s: Adding case #%s" % (UNASSIGNED, case["id"]))
            parsed_cases[UNASSIGNED].append(case)

    return parsed_cases


def create_emails(
    subject: str,
    message: str,
    parsed_cases: dict,
    notify: list,
    base_url: str,
    search_criteria=None,
    dry: bool = False,
    from_address: str = "argus-noreply@mnemonic.no",
) -> dict:
    """Crates emails from the parsed_cases dict

    :param subject: The subject of the email
    :param message: The boilerplate part of the email
    :param parsed_cases: The parsed cases to create emails from
    :param notify: Who to notify for unassigned cases
    :param search_criteria: dict of search criterias used to link to a relevant
      search. Used only if there are more than one case.
    :param from_address: The sender address to use
    :return: A mail for each user
    """
    mails = {}

    for user, cases in parsed_cases.items():
        if len(cases) == 0:
            # Don't bother generating a mail without any cases
            continue

        body = ""

        # generate UI search link for multiple cases
        if search_criteria is not None and len(cases) > 1:
            search_url = f"https://{base_url}/spa/case/filter?filterSearch={quote(json.dumps(search_criteria))}"
            body += f'<p><a href="{search_url}">all mentioned cases</a></p>'

        body += f"<p>{message}</p>\n<br>\n"
        for case in cases:
            body += (
                '<p><a href="https://{base_url}/spa/case/view/{case_id}">#{case_id}</a> '
                "- {status} - {subject} - Last Update: {time}</p>\n".format(
                    base_url=base_url,
                    case_id=case["id"],
                    status=case["status"],
                    subject=case["subject"],
                    time=datetime.fromtimestamp(
                        case["lastUpdatedTimestamp"] / 1000
                    ).strftime("%Y-%m-%d %H:%M:%S"),
                )
            )

        mails[user] = MIMEText(body, "html")
        mails[user]["Subject"] = subject
        mails[user]["From"] = from_address
        mails[user]["To"] = (
            ";".join(notify) if user == UNASSIGNED else "%s@mnemonic.no" % user
        )

        if dry:
            print(f'To: {mails[user]["To"]}')
            print(f'From: {mails[user]["From"]}')
            print(f'Subject: {mails[user]["Subject"]}')
            print(body)
            print("------------------------------")

    return mails


@register_command(extending="cases", module=argus_cli_module)
def remind(
    subject: str,
    message: str,
    notify: list = None,
    case_type: utils.CASE_TYPES = None,
    service: str = None,
    status: utils.STATUSES = DEFAULT_STATUS,
    customer: str = None,
    exclude_customer: get_customer_id = None,
    priority: utils.PRIORITIES = None,
    skip: SKIPABLES = [],
    skip_assigned: bool = False,
    skip_unassigned: bool = False,
    skip_notify_tech: bool = False,
    skip_not_due: bool = False,
    start: date_or_relative = None,
    end: date_or_relative = None,
    days: int = 14,
    dry: bool = False,
    from_address: str = "argus-noreply@mnemonic.no",
    smtp_host: str = "smtp.mnemonic.no",
    base_url: str = "argusweb.mnemonic.no",
):
    """A command for reminding people when they have pending cases that haven't been updated for a while.

    :param subject: The subject of the email
    :param message: The body of the email
    :param notify: Email(s) to notify for unassigned cases
    :param list case_type: The log type of the case
    :param list service: The service type of the case
    :param list status: The status of the case
    :param list customer: Customers to use (shortname)
    :param list exclude_customer: Customers to exclude
    :param list priority: Priorities to have on the case
    :param list skip: Certain things to not notify about -- DEPRECATED
    :param skip_assigned: Exclude cases which have a tech assignee
    :param skip_not_due: Exclude cases which have not reached the tech due date
    :param skip_unassigned: Exclude cases which are not assigned to any tech
    :param skip_notify_tech: Skip sending notifications to assigned tech, instead only send to notify list (unassigned cases will always be sent to notify list)
    :param start: Only consider cases created after this date
    :param days: Only consider cases updated within the last X days
    :param dry: Runs the program without sending the actual email
    :param from_address: The from-address to send mail from
    :param smtp_host: The SMTP host to send mail from
    :param base_url: The base URL for formatting the email link

    :alias smtp_host: smtp
    """

    if "unassigned" in skip:
        skip_unassigned = True

    if "assigned" in skip:
        skip_assigned = True

    if "tech" in skip:
        skip_notify_tech = True

    log.info("Getting cases to notify about")

    search_criteria = {
        "limit": 0,
        "type": case_type,
        "service": service,
        "status": status,
        "priority": priority,
        "customer": customer,
        "startTimestamp": str(start) if start is not None else None,
        "timeFieldStrategy": ["createdTimestamp"],
        "subCriteria": [],
    }

    if days:
        search_criteria["subCriteria"].append(
            {
                "required": True,
                "timeFieldStrategy": ["lastUpdatedTimestamp"],
                "endTimestamp": ("-%dd" % days),
            }
        )

    if skip_not_due:
        search_criteria["subCriteria"].append(
            {
                "exclude": True,
                "timeFieldStrategy": ["techDueTimestamp"],
                "startTimestamp": "now",
            }
        )

    if skip_assigned:
        search_criteria["subCriteria"].append(
            {
                "required": True,
                "techAssigned": False,
            }
        )

    if skip_unassigned:
        search_criteria["subCriteria"].append(
            {
                "required": True,
                "techAssigned": True,
            }
        )

    if exclude_customer:
        search_criteria["subCriteria"].append(
            {"exclude": True, "customerID": exclude_customer or None}
        )

    result = advanced_case_search(**search_criteria)

    cases = result["data"]
    parsed_cases = parse_data(cases, skip_notify_tech)

    if parsed_cases[UNASSIGNED] and not notify:
        print("No --notify recipient specified")
        sys.exit(1)

    log.info("Creating emails")
    mails = create_emails(
        subject,
        message,
        parsed_cases,
        notify or [],
        base_url,
        dry=dry,
        search_criteria=search_criteria,
        from_address=from_address,
    )

    if len(mails) == 0:
        log.info("No emails to send")
    elif dry:
        log.info("Not sending any emails (DRY RUN)")
    else:
        log.info("Sending email")
        for recipient, mail in mails.items():
            smtp = smtplib.SMTP(smtp_host)
            smtp.send_message(mail)
            smtp.quit()


if __name__ == "__main__":
    run(remind)



import collections
from argparse import ArgumentParser


class ImmutableDeepDict(collections.abc.MutableMapping):
    """A dictionary with ability for deep get and set

    A key turns immutable after being set.
    """

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)

    def __getitem__(self, keys):
        """Gets a plugin

        :param keys: The name and hierarchy of the plugin
        :returns: The dict that holds the plugins commands
        """

        def recurse(name: tuple, parent: dict) -> dict:
            key = name[0]

            if key not in parent:
                raise KeyError("The key does not exist!")
            elif len(name) > 1:
                return recurse(name[1:], parent[key])
            else:
                return parent[key]

        if not keys:
            raise KeyError("Key is empty")
        elif isinstance(keys, str):
            keys = (keys,)
        return recurse(keys, self._dict)

    def __setitem__(self, keys, value):
        """Adds a key

        Fills missing steps with a empty dict

        :param keys: The name and hierarchy of the plugin
        :returns: The dict that holds the plugins commands
        """

        def recurse(name: tuple, parent: dict) -> dict:
            key = name[0]

            if len(name) > 1:
                if key not in parent:
                    parent[key] = {}
                return recurse(name[1:], parent[key])
            elif key in parent:
                raise KeyError("The key already exists")
            else:
                parent[key] = value
                return parent[key]

        if not keys:
            raise KeyError("Key is empty")
        elif isinstance(keys, str):
            keys = (keys,)
        return recurse(keys, self._dict)

    def __delitem__(self, key):
        del self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __str__(self):
        return str(self._dict)


def set_dotnotation(original: dict, dotnotation: str, value=None) -> dict:
    """Use dotted notation to create set values in a dict,
    e.g:
    >>> dotattr({}, 'some.path.in.dict', 'value')
    { "some": { "path": { "in": {"dict": "value" }}}}

    >>> dotattr({'credentials': {'username': 'test'}}, 'credentials.username')
    'test'
    """
    original = original or {}
    parts = dotnotation.split(".")
    current = original
    for part in parts:
        if part is parts[-1]:
            current[part] = value
        elif part not in current:
            current[part] = {}
        current = current[part]
    return original


def get_dotnotation(original: dict, dotnotation: str) -> dict:
    """Use dotted notation to create set values in a dict,
    e.g:
    >>> dotattr({}, 'some.path.in.dict', 'value')
    { "some": { "path": { "in": {"dict": "value" }}}}

    >>> dotattr({'credentials': {'username': 'test'}}, 'credentials.username')
    'test'
    """
    original = original or {}
    parts = dotnotation.split(".")
    current = original
    for part in parts:
        if part is parts[-1]:
            return current[part]
        elif part not in current:
            current[part] = {}
        current = current[part]
    return original



from collections import defaultdict

NOT_AVAILABLE_TEXT = ""


class _defaultdict(defaultdict):
    """This class is introduced because by default default dicts will print as a class, not a dict."""

    def __str__(self):
        return str(dict(self))

    def __repr__(self):
        return str(dict(self))


def _dict_to_default_dict(d: dict) -> dict:
    """Converts a dict to a default dict.

    This is done to convert dicts to

    :param d: A dictionary with data
    :return: A dictionary that will return "not available" when accessing non-existing indices.
    """
    if isinstance(d, dict):
        for key, value in d.items():
            d[key] = _dict_to_default_dict(value)
        d = _defaultdict(lambda: NOT_AVAILABLE_TEXT, **d)
    return d


def formatted(
    data: list, formatter: callable, headers: list = None, show_headers: bool = True
) -> str:
    """Helper for creating formatted outputs

    :param data: Data to output
    :param formatter: Function that turns a list of data into a string based on the headers
    :param headers: Headers of the data, defaults to the keys of the first piece of data.
    :param show_headers: Whether or not to display the headers
    :returns: A formatted string based on the input-data
    """
    if not headers:
        try:
            headers = data[0].keys()
        except IndexError:
            # If there is no data
            headers = []

    out = []

    if show_headers:
        out.append(
            formatter({header: header for header in headers}, headers, is_header=True)
        )

    for row in data:
        row = _dict_to_default_dict(row)
        out.append(formatter(row, headers))

    return "\n".join(out)


def _flatten_data(row: dict, headers: list):
    """Flatten the data and put it in the same order as the headers."""
    if isinstance(row, dict):
        return [str(row[field]) for field in headers]
    else:
        # Assumes that if it isn't dict, it's a class with attrs
        return ([str(getattr(row, field)) for field in headers],)


def jira_table(data: list, **kwargs):
    """Outputs a JIRA table

    :param data: A list of dicts to format.
    :param kwargs: Parameters passed to `formatted`.
    """

    def table_formatter(row: dict, headers: list, is_header=False, **kwargs):
        seperator = "||" if is_header else "|"
        return seperator + seperator.join(_flatten_data(row, headers)) + seperator

    return formatted(data, table_formatter, **kwargs)


def csv(data: list, **kwargs):
    """Standard CSV printer

    :param data: A list of dicts to format.
    :param kwargs: Parameters passed to `formatted`.
    """

    def csv_formatter(row: dict, headers: list, **kwargs):
        return ",".join(_flatten_data(row, headers))

    return formatted(data, csv_formatter, **kwargs)


def formatted_string(data: list, format_string: str, **kwargs) -> str:
    """Using format-strings for outputs.

    Simply calls .format(...) on the supplied string

    :param data: A list of dicts to format.
    :param format_string: The string to use when outputting.
    :param kwargs: Parameters passed to `formatted`.
    """
    if "show_headers" not in kwargs:
        kwargs["show_headers"] = False

    def string_formatter(row: dict, headers: list, **kwargs):
        return format_string.format(**row)

    return formatted(data, string_formatter, **kwargs)



import sys
from argparse import ArgumentTypeError
from datetime import datetime
from enum import Enum
from io import StringIO
from pathlib import Path

import dateparser
import yaml
from argus_cli.helpers.log import log
from argus_cli.plugin import register_command
from argus_cli.utils.formatting import get_data_formatter
from argus_cli.utils.time import time_parser
from argus_plugins import argus_cli_module
from argus_plugins.cases.autocreate_case import CLOSED_CASE_UPDATE_METHODS, autocreate
from argus_plugins.cases.utils import CASE_TYPES, PRIORITIES, STATUSES
from argus_plugins.events.search import search
from argus_plugins.events.search_model import SearchOptions
from pydantic import BaseModel, ValidationError, validator
from pydantic.typing import Dict, List, Literal, Optional, Union

# convert lists of values to enums for validation
StatusEnum = Enum("StatusEnum", {v: v for v in STATUSES})
PrioritiesEnum = Enum("PrioritiesEnum", {v: v for v in PRIORITIES})
CaseTypesEnum = Enum("CaseTypesEnum", {v: v for v in CASE_TYPES})
ClosedCaseUpdateMethodEnum = Enum(
    "ClosedCaseUpdateMethod",
    {v: v for v in CLOSED_CASE_UPDATE_METHODS},
)


def _relative_time(value: str, base: Optional[datetime] = None) -> datetime:
    """parses a string as a relative datetime.

    if ``base`` is set, input will be parsed as relative to ``base``, otherwise
    it will be parsed as relative to the current datetime.

     :param value: time string to parse. "ago" will be appended to this value if not
       already present. Example: "1 hour" will be parsed as "1 hour ago".
    :param base: if set, point in time to which the value is relative.
    """
    altered_value = value.replace("ago", "").strip() + " ago"
    parse_settings = {"RELATIVE_BASE": base} if base else None
    parsed = dateparser.parse(altered_value, settings=parse_settings)
    if parsed is None:
        raise ValueError(
            f'could not parse time string "{value}" (parsed as: "{altered_value}")'
        )
    return parsed


class TitleFormat(str, Enum):
    fmtstr = ("fmtstr",)
    jinja = "jinja"


class LocalizedCaseTitle(BaseModel):
    en: str
    no: str

    def to_args(self):
        return {"case_title": None, "case_title_no": self.no, "case_title_en": self.en}


class AutocreateCaseOptions(BaseModel):
    group_by: List[str] = []
    title: Optional[Union[str, LocalizedCaseTitle]] = None
    title_format: TitleFormat = TitleFormat.fmtstr
    status: StatusEnum = StatusEnum.pendingCustomer
    priority: PrioritiesEnum = PrioritiesEnum.medium
    type: StatusEnum = CaseTypesEnum.securityIncident
    service: str = "ids"
    category: Optional[str] = None
    close_after_create: bool = False
    use_fields: bool = False

    tags: Optional[Dict[str, str]] = None
    fields: Optional[Dict[str, str]] = None

    class Config:
        use_enum_values = True

    def to_args(self) -> dict:
        """convert the model instance to a dict of arguments ready to be passed
        to the ``autocreate`` function."""
        title_opts = {"case_title": None, "case_title_no": None, "case_title_en": None}
        if isinstance(self.title, str):
            title_opts["case_title"] = self.title
        elif isinstance(self.title, LocalizedCaseTitle):
            title_opts = self.title.to_args()

        return {
            **title_opts,
            "case_status": self.status,
            "case_priority": self.priority,
            "case_type": self.type,
            "case_title_format": self.title_format,
            "case_service": self.service,
            "case_category": self.category,
            "tags": self.tags,
            "fields": self.fields,
            "close_after_create": self.close_after_create,
            "group_by": self.group_by,
            "use_fields": self.use_fields,
        }


class AutocreateNotificationOptions(BaseModel):
    skip_notifications: bool = False
    watcher: Optional[str] = None
    watcher_from_field: Optional[str] = None
    watcher_verbose: Optional[bool] = False

    def to_args(self):
        """convert the model instance to a dict of arguments ready to be passed
        to the ``autocreate`` function."""
        return {
            "case_watcher": self.watcher,
            "case_watcher_from_field": self.watcher_from_field,
            "skip_notifications": self.skip_notifications,
            "case_watcher_verbose": self.watcher_verbose,
        }


class AutocreateWorkflowsOptions(BaseModel):
    request_soc_analysis: Optional[bool] = False
    request: Optional[str] = None
    acknowledge: Optional[str] = None
    comment: Optional[str] = None

    def to_args(self):
        """convert the model instance to a dict of arguments ready to be passed
        to the ``autocreate`` function."""
        return {
            "request_soc_analysis": self.request_soc_analysis,
            "request_workflow": self.request,
            "acknowledge_workflow": self.acknowledge,
            "workflow_comment": self.comment,
        }


class AutocreateUpdateOptions(BaseModel):
    closed_case: Optional[ClosedCaseUpdateMethodEnum] = None
    silent: bool = False
    status_on_update: Optional[StatusEnum] = None

    def to_args(self):
        """convert the model instance to a dict of arguments ready to be passed
        to the ``autocreate`` function."""
        return {
            "closed_case_update": self.closed_case,
            "status_on_update": self.status_on_update,
            "silent_update": self.silent,
        }


class AutocreateOptions(BaseModel):
    """Configuration model for the ``autocreate-yaml`` command.

    Allows representing arguments to both ``events search`` and ``cases autocreate``
    commands.
    """

    autocreate_yaml_version: Literal[1] = 1
    search: SearchOptions

    key: str
    template_folder: Optional[Path] = None

    base_time: str = "now"
    time_frame: str
    timeout: Optional[str] = None

    sort_by: List[str] = []

    case: AutocreateCaseOptions = AutocreateCaseOptions()
    notifications: AutocreateNotificationOptions = AutocreateNotificationOptions()
    workflows: AutocreateWorkflowsOptions = AutocreateWorkflowsOptions()
    updates: AutocreateUpdateOptions = AutocreateUpdateOptions()

    initial_internal_comment: bool = False
    internal_case: bool = False
    send_to_qa: bool = False
    explicit_access: List[str] = []

    attach_events: Optional[Literal["json", "csv"]] = None

    test_data: bool = False
    default_prod_excluded_flags: List[str] = ["NOTIFIED", "INITIAL_TUNING"]

    class Config:
        use_enum_values = True

    def search_time_args(self) -> dict:
        """converts time values to a dict of arguments ready to be passed to the
        events ``search`` function.
        """
        return {
            "start": int(self.start_dt.timestamp() * 1000),
            "end": int(self.end_dt.timestamp() * 1000),
        }

    @property
    def end_dt(self):
        return _relative_time(self.base_time)

    @property
    def start_dt(self):
        return _relative_time(self.time_frame, base=self.end_dt)

    @property
    def timeout_dt(self):
        if not self.timeout:
            return datetime.now()
        return _relative_time(self.timeout, base=self.end_dt)

    def to_search_args(self):
        """converts relevant settings to the full set of arguments expected by
        the events ``search`` function.
        """
        return {
            **self.search.to_search_args(),
            **self.search_time_args(),
            "format": get_data_formatter("json"),
            "print_events": False,
        }

    def to_autocreate_args(self):
        """converts relevant settings to the full set of arguments expected by
        the events ``autocreate`` function.
        """
        args = self.dict(
            exclude={
                "search",
                "case",
                "notifications",
                "workflows",
                "updates",
                "base_time",
                "time_frame",
                "default_prod_excluded_flags",
                "timeout",
                "autocreate_yaml_version",
            }
        )
        args.update(
            {
                **self.notifications.to_args(),
                **self.case.to_args(),
                **self.notifications.to_args(),
                **self.workflows.to_args(),
                **self.updates.to_args(),
                "timeout": self.timeout_dt,
            }
        )
        return args

    @validator("base_time", always=True)
    def validate_datetime(cls, v):
        try:
            time_parser(v)
        except ArgumentTypeError:
            raise ValueError(f'could not parse time string "{v}"')
        return v

    @validator("time_frame", "timeout", always=True)
    def validate_relative_dt(cls, v):
        if v is None:
            return v
        _relative_time(v)
        return v


def debug_time_settings(
    settings: AutocreateOptions,
) -> None:
    """print out information about time configuration parsing.

    :param settings: autocreate configuration.
    """
    base_time_dt = settings.end_dt
    base_time_ts = int(base_time_dt.timestamp() * 1000)
    start_time_dt = settings.start_dt
    start_time_ts = int(start_time_dt.timestamp() * 1000)

    print("SEARCH INTERVAL:")
    print(
        f'base_time (end time): "{settings.base_time}" - parsed: "{base_time_dt}" ( timestamp: {base_time_ts} )'
    )
    print(f"start time: {start_time_dt} ( timestamp: {start_time_ts} )")
    print(
        f'start time calculated based on time_frame "{settings.time_frame}" parsed as "{settings.time_frame} ago" from base_time ({base_time_dt})'
    )
    print(
        f'search interval: from "{start_time_dt}" to "{base_time_dt} ({start_time_ts} to {base_time_ts})"'
    )
    print()
    print("TIMEOUT:")
    if settings.timeout is None:
        print(f'timeout: not set, using current datetime ("{settings.timeout_dt}")')
    else:
        timeout_dt = settings.timeout_dt
        timeout_ts = int(timeout_dt.timestamp() * 1000)
        print(
            f'timeout: "{settings.timeout}" parsed as "{settings.timeout} ago" from base_time ("{base_time_dt}") - parsed: "{timeout_dt}" ( timestamp: {timeout_ts} )'
        )


@register_command(extending="cases", module=argus_cli_module)
def autocreate_yaml(
    config: Path,
    template_folder: Path = None,
    dry: bool = False,
    base_time: str = None,
    time_frame: str = None,
    test_time_expr: bool = False,
    validate: bool = False,
    prod_excludes: bool = True,
):
    """A tool for automatically creating a case based on event - YAML edition

    Values from the configuration file will be passed to the "events search"
    and "cases autocreate" commands.

    Example configuration:
        key: "example"
        template_folder: "test_templates"
        time_frame: "1 day"
        timeout: "1 hour"
        search:
          customer:
            include:
              - "mnemonic"

    :param config: path to the autoreport configuration file.
    :param dry: if set, no data will be commited by autocreate.
    :param base_time: if set, overrides the base_time set in the configuration.
    :param time_frame: if set, overrides the time_frame set in the configuration.
    :param test_time_expr: if set, print information about time values parsing and exit.
    :param validate: if set, only check that the configuration is valid and exit.
    :param prod_excludes: if set to false with --no-prod-excludes, flags that are
      excluded by default during event search will not be excluded.
    """
    settings_raw = yaml.safe_load(config.expanduser().read_text())

    if validate:
        try:
            AutocreateOptions.validate(settings_raw)
        except ValidationError as e:
            print(str(e))
            sys.exit(1)
        print(f'"{config}" contains valid autocreate configuration data.')
        sys.exit(0)

    settings = AutocreateOptions(**settings_raw)
    if template_folder:
        settings.template_folder = template_folder
    elif not settings.template_folder:
        print(
            "no template_folder is set in the configuration file, --template-folder is required."
        )
        exit(1)

    if base_time:
        settings.base_time = base_time
    if time_frame:
        settings.time_frame = time_frame

    if test_time_expr:
        debug_time_settings(settings)
        sys.exit()

    search_args = settings.to_search_args()

    if prod_excludes:
        log.info(
            f"adding production excluded flags: {settings.default_prod_excluded_flags}"
        )
        search_args["exclude_flag"] = (
            search_args.get("exclude_flag") or []
        ) + settings.default_prod_excluded_flags

    log.info("running search")
    search_results = search(**search_args)
    search_output = StringIO()
    search_output.write(search_results)
    search_output.seek(0)

    autocreate_args = settings.to_autocreate_args()
    autocreate_args["data"] = search_output
    if dry:
        autocreate_args["dry"] = True

    log.info("running autocreate")
    autocreate(**autocreate_args)



from enum import Enum

from argus_api.exceptions.http import ObjectNotFoundException
from argus_plugins.cases.utils import get_customer_id
from argus_plugins.enrichments import EVENTS_ENRICHMENTS
from argus_plugins.events.utils import FLAGS, SEVERITIES
from pydantic import BaseModel, validator
from pydantic.typing import Dict, List, Optional, Union

FlagEnum = Enum("FlagEnum", {f: f for f in FLAGS})
SeverityEnum = Enum("SeverityEnum", {s: s for s in SEVERITIES})
EnrichEnum = Enum("EnrichEnum", {e: e for e in EVENTS_ENRICHMENTS})


class IncludeExcludeStrings(BaseModel):
    include: Optional[List[str]] = None
    exclude: Optional[List[str]] = None


class IncludeExcludeInts(BaseModel):
    include: Optional[List[int]] = None
    exclude: Optional[List[int]] = None


class IncludeExcludeCustomers(IncludeExcludeStrings):
    include: Optional[List[Union[str, int]]] = None
    exclude: Optional[List[Union[str, int]]] = None

    @validator("include", "exclude", each_item=True)
    def get_customer_id(cls, v):
        if v is None:
            return None
        try:
            return get_customer_id(v)
        except ObjectNotFoundException:
            raise ValueError(f'customer "{v}" not found')


class IncludeExcludeProps(BaseModel):
    include: Optional[Dict[str, str]] = None
    exclude: Optional[Dict[str, str]] = None


class IPSearchOptions(BaseModel):
    any: Optional[IncludeExcludeStrings] = None
    source: Optional[IncludeExcludeStrings] = None
    destination: Optional[IncludeExcludeStrings] = None


class SearchOptions(BaseModel):
    """pydantic model of the events search command options."""

    customer: Optional[IncludeExcludeCustomers] = None
    flag: Optional[IncludeExcludeStrings] = None
    alarm: Optional[IncludeExcludeStrings] = None
    signature: Optional[IncludeExcludeStrings] = None
    properties: Optional[IncludeExcludeProps] = None
    attack_category_id: Optional[IncludeExcludeInts] = None
    ip: Optional[IncludeExcludeStrings] = None
    source_ip: Optional[IncludeExcludeStrings] = None
    destination_ip: Optional[IncludeExcludeStrings] = None

    exact_match_properties: bool = True
    min_severity: SeverityEnum = SeverityEnum.high
    min_count: Optional[int] = None
    source_ip_min_bits: Optional[int] = None
    destination_ip_min_bits: Optional[int] = None
    enrich: Optional[List[EnrichEnum]] = None

    class Config:
        use_enum_values = True

    def get_include_arg(
        self,
        arg_name,
    ):

        field = getattr(self, arg_name, None)
        return {
            f"include_{arg_name}": field.include if field else None,
            f"exclude_{arg_name}": field.exclude if field else None,
        }

    def to_search_args(self):

        include_exclude_fields = (
            "properties",
            "flag",
            "alarm",
            "signature",
            "attack_category_id",
            "ip",
            "source_ip",
            "destination_ip",
            "customer",
        )
        args = self.dict(exclude=set(include_exclude_fields))
        for thing in include_exclude_fields:
            args.update(self.get_include_arg(thing))
        return args



import logging
import logging.config

# Set up custom log levels for plugins
logging.PLUGIN = 21


class ArgusCLILog(logging.getLoggerClass()):
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

        logging.addLevelName(logging.PLUGIN, "PLUGIN")

    def plugin(self, msg, *args, **kwargs):
        """Custom log level for plugins"""
        if self.isEnabledFor(logging.PLUGIN):
            self._log(logging.PLUGIN, msg, args, **kwargs)


def setup_logger(settings: dict):
    logging.setLoggerClass(ArgusCLILog)
    logging.config.dictConfig(settings)


# The logger!
log = logging.getLogger("argus_cli")
log.propagate = False
